unit SControls;

{$mode objfpc}{$H+}
{$R shell.res}

interface

uses
  Classes, SysUtils, Controls, LCLIntf, Graphics, SForms, SThemes,
  Forms, Math, ImgList, LCLType, lazutf8, Dialogs;

type

  { TStatusBar }

  TStatusAlign = (saLeft,saCenter,saRight);
  TStatusItem = class
    left    : integer;
    align   : TStatusAlign;
  end;

  TStatusBar = class(TCustomControl)
  protected
    procedure MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure ThisCursor(X,Y: Integer);
  private
    { private declarations }
    fdown   : boolean;
    fmouse  : TPoint;
    fbounds : TRect;
    fitems  : TStringList;
  public
    { public declarations }
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Resize; override;
    procedure   Add(const txt: string; const x: Integer; const a: TStatusAlign);
    procedure   Item(const n: integer; const txt: string);
    procedure   Paint; override;
  end;

  TScrollBarKind = Forms.TScrollBarKind;

  {TSScrollBar}

  TSScrollBar = class(TCustomControl)
  protected
    procedure   MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure   MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   Resize; override;
    procedure   Paint; override;
  private
    fPix        : integer;
    fKind       : TScrollBarKind;
    fPos        : Integer;
    fMin        : Integer;
    fMax        : Integer;
    fPage       : Integer;
    fdown       : boolean;
    fmouse      : TPoint;
    fthumb      : TRect;
    fbounds     : TRect;
    fshow       : boolean;
    fshift      : integer;
    fOnChange   : TNotifyEvent;
    procedure   SetKind(AValue: TScrollBarKind);
    procedure   SetMax(AValue: Integer);
    procedure   SetMin(AValue: Integer);
    procedure   SetPage(AValue: Integer);
    procedure   SetPos(AValue: Integer);
    procedure   SetShift(AValue: integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    property    Kind: TScrollBarKind read fKind write SetKind;
    property    Min: Integer read fMin write SetMin;
    property    Max: Integer read fMax write SetMax;
    property    Position: Integer read fPos write SetPos;
    property    PageSize: Integer read fPage write SetPage;
    property    Shift: integer read fShift write SetShift;
    property    OnChange: TNotifyEvent read fOnChange write fOnChange;
  published

  end;

  { TSButton }

  TSButton = class(TCustomControl)
  protected
    procedure MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure Click; override;
    procedure MouseEnter; override;
    procedure MouseLeave; override;
  private
    { private declarations }
    fdown    : boolean;
    fenter   : boolean;
    fonclick : TNotifyEvent;
  public
    { public declarations }
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Paint; override;
    property    OnClick: TNotifyEvent read fonclick write fonclick;
  end;

  { TSShellView }

  TShellItem = class
    attr     : integer;
    path     : string;
    ext      : string;
    selected : boolean;
    isempty  : boolean;
  end;

  TSShellView = class(TCustomControl)
  protected
    procedure   DblClick; override;
    procedure   MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure   MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   Resize; override;
    procedure   Paint; override;
  private
    fdown       : boolean;
    fstart      : TPoint;
    fend        : TPoint;
    files       : TStringList;
    fvscroll    : TSScrollBar;
    fimages     : TCustomImageList;
    fselect     : TRect;
    fmultisel   : boolean;
    fheader     : integer;
    fdata       : TStringList;
    fondblclick : TNotifyEvent;
    fonselect   : TNotifyEvent;
    procedure   CheckSelected;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   ScanFile(str: string);
    property    OnDblClick  : TNotifyEvent read fondblclick write fondblclick;
    property    OnSelect    : TNotifyEvent read fonselect write fonselect;
    property    FileData    : TStringList read fdata;
    property    MultiSelect : boolean read fmultisel write fmultisel;
  published

  end;

implementation

{ TSButton }

constructor TSButton.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  color := themes.buttons.color;
  width := 80;
  height := 27;
  caption := 'Button';
  font.size := themes.buttons.font.size;
  font.color := themes.buttons.font.color;
end;

destructor TSButton.Destroy;
begin

  inherited Destroy;
end;

procedure TSButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  fdown := Button = mbLeft;
  if fdown then color := themes.buttons.light;
end;

procedure TSButton.MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  color := themes.buttons.enter;
  fdown := false;
end;

procedure TSButton.Click;
begin
  inherited Click;
  fdown := false;
  if assigned(fonclick) then fonclick(self);
end;

procedure TSButton.MouseEnter;
begin
  inherited MouseEnter;
  color  := themes.buttons.enter;
end;

procedure TSButton.MouseLeave;
begin
  inherited MouseLeave;
  color  := themes.buttons.color;
end;

procedure TSButton.Paint;
var
  tw,th: integer;
begin
  inherited Paint;
  tw := canvas.textwidth(caption);
  th := canvas.textheight(caption);
  canvas.textout((width-tw) div 2,(height-th) div 2,caption);
end;

{ TStatusBar }

constructor TStatusBar.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  color  := themes.status.color;
  align  := alBottom;
  height := 25;
  fitems := TStringList.Create;
  font.Size  := 9;
  font.color := themes.form.font.color;
  if TheOwner is TSForm then TSForm(TheOwner).sbheight:=height;
end;

destructor TStatusBar.Destroy;
var
  i: integer;
begin
  for i:=0 to fitems.count -1 do fitems.objects[i].free;
  FreeAndNil(fitems);
  inherited Destroy;
end;

procedure TStatusBar.ThisCursor(X,Y: Integer);
begin
  if      PtInRect(Rect(0,0,themes.form.border.size,height-themes.form.border.size-5),Point(X,Y)) then            cursor := crSizeE
  else if PtInRect(Rect(width-themes.form.border.size,0,width,height-themes.form.border.size-5),Point(X,Y)) then  cursor := crSizeW
  else if PtInRect(Rect(width-themes.form.border.size-5,height-themes.form.border.size-5,width,height),Point(X,Y)) then                    cursor := crSizeNW
  else if PtInRect(Rect(themes.form.border.size+5,height-themes.form.border.size,width-themes.form.border.size-5,height),Point(X,Y)) then cursor := crSizeS
  else if PtInRect(Rect(0,height-themes.form.border.size-5,themes.form.border.size+5,height),Point(X,Y)) then                              cursor := crSizeNE
  else cursor := crDefault;
end;

procedure TStatusBar.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  fdown   := Button = mbLeft;
  fbounds := TSForm(parent).BoundsRect;
  GetCursorPos(fmouse);
end;

procedure TStatusBar.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  p       : TPoint;
  l,t,w,h : integer;
begin
  inherited MouseMove(Shift, X, Y);
  if not fdown then
    ThisCursor(X,Y)
  else begin
    GetCursorPos(p);
    l := TSForm(parent).left;
    t := TSForm(parent).top;
    w := TSForm(parent).width;
    h := TSForm(parent).height;

    case cursor of
        crSizeE  :
                   begin
                     l := fbounds.left +  p.x - fmouse.x;
                     w := (fbounds.right-fbounds.left) - (p.x - fmouse.x);
                   end;
        crSizeW  : w := (fbounds.right-fbounds.left) + (p.x - fmouse.x);
        crSizeS  : h := (fbounds.bottom-fbounds.top) + (p.y - fmouse.y);

        crSizeNW :
                   begin
                     w := (fbounds.right-fbounds.left) + (p.x - fmouse.x);
                     h := (fbounds.bottom-fbounds.top) + (p.y - fmouse.y);
                   end;

        crSizeNE :
                   begin
                     h := (fbounds.bottom-fbounds.top) + (p.y - fmouse.y);
                     l := fbounds.left +  p.x - fmouse.x;
                     w := (fbounds.right-fbounds.left) - (p.x - fmouse.x);
                   end;
      end;
      TSForm(parent).SetBounds(l,t,w,h);
      {$IFDEF WINDOWS}TSForm(parent).Update;{$ENDIF}
  end;
end;

procedure TStatusBar.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  fdown   := false;
  ThisCursor(X,Y);
end;

procedure TStatusBar.Resize;
begin
  inherited Resize;

end;

procedure TStatusBar.Add(const txt: string; const x: Integer; const a: TStatusAlign);
var
  si : TStatusItem;
begin
  si := TStatusItem.Create;
  si.left  := x;
  si.align := a;
  fitems.AddObject(txt,si);
end;

procedure TStatusBar.Item(const n: integer; const txt: string);
begin
  if (n >= 0 ) and (n < fitems.count) then begin
    fitems.strings[n] := txt;
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TStatusBar.Paint;
var
  i,tw,w,x: integer;
begin
  inherited Paint;
  canvas.pen.width := 1;
  canvas.pen.color := themes.form.border.color;
  canvas.line(0,0,0,height);
  canvas.line(0,height-1,width-1,height-1);
  canvas.line(width-1,height-1,width-1,0);

  canvas.pen.color := themes.status.bgcolor;
  canvas.brush.color := themes.status.bgcolor;
  canvas.Polygon([Point(width-5-4,height-4),Point(width-4,height-4),Point(width-4,height-4-5)]);
  canvas.line(width-5-7,height-4,width-3,height-8-5);

  canvas.brush.style := bsClear;
  for i:=0 to fitems.count - 1 do begin
    x := TStatusItem(fitems.objects[i]).left;
    tw := canvas.TextWidth(fitems.strings[i]);
    if i < fitems.count - 1 then
      w :=  TStatusItem(fitems.objects[i+1]).left-TStatusItem(fitems.objects[i]).left
    else
      w := width - TStatusItem(fitems.objects[i]).left;
    case TStatusItem(fitems.objects[i]).align of
      saLeft: x := x + 10;
      saCenter: x := x + (w-tw) div 2;
      saRight:;
    end;
    canvas.pen.width := 2;
    canvas.pen.color := themes.status.separ;
    if i>0 then canvas.line(TStatusItem(fitems.objects[i]).left,4,TStatusItem(fitems.objects[i]).left,height-6);
    canvas.TextOut(x,(height-canvas.TextHeight('X')) div 2 - 1,fitems.Strings[i]);
  end;
end;

{ TSScrollBar }

constructor TSScrollBar.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  color  := themes.scroll.color;
  width  := 14;
  height := 14;
  fKind  := sbHorizontal;
  fMin   := 0;
  fMax   := 100;
  fPage  := 20;
  fthumb := Rect(0,0,width,height);
  fshow  := false;
  fshift := 0;
end;

destructor TSScrollBar.Destroy;
begin

  inherited Destroy;
end;

procedure TSScrollBar.SetShift(AValue: integer);
begin
  if fShift=AValue then Exit;
  fShift:=AValue;
  if handleallocated then InvalidateRect(handle,nil,false);
end;

procedure TSScrollBar.SetKind(AValue: TScrollBarKind);
begin
  if fKind <> AValue then begin
    fKind := AValue;
    if handleallocated then InvalidateRect(handle,nil,false);
  end;
end;

procedure TSScrollBar.SetMax(AValue: Integer);
begin
  if (AValue < 0) then AValue := 0;
  if fMax<>AValue then begin
    fMax:= AValue;
    fPos:= Math.Min(fPos, fMax);
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSScrollBar.SetMin(AValue: Integer);
begin
  if fMin<>AValue then begin
    fMin:= AValue;
    fPos:= Math.Max(fPos, fMin);
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSScrollBar.SetPage(AValue: Integer);
begin
  if fPage <> AValue then fPage := AValue;
  if fKind = sbVertical then
    fthumb.bottom := round(fthumb.top + fPage * height/(fMax+1))
  else
    fthumb.right := round(fthumb.left + fPage * (width-fshift)/(fMax+1));
  if fMax < fPage then
    fshow := false
  else begin
    fshow := true;
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSScrollBar.SetPos(AValue: Integer);
var
  w: integer;
  found: boolean;
begin
  if AValue = -1 then found := true;
  AValue:= Math.Min(AValue, fMax);
  AValue:= Math.Max(AValue, fMin);
  if found or (fPos <> AValue) then begin
    fPos := AValue;
    if fKind = sbVertical then begin
      w := fthumb.bottom - fthumb.top;
      fthumb.top := round(fPos * height/(fMax+1));
      fthumb.bottom := fthumb.top + w;
      if fthumb.top < 0 then begin
        fthumb.top := 0;
        fthumb.bottom := fthumb.top + w;
      end;
      if fthumb.bottom > height then begin
        fthumb.bottom := height;
        fthumb.top  := fthumb.bottom - w;
      end;
    end else begin
      w := fthumb.right - fthumb.left;
      fthumb.left  := round((fPos-1) * ((width-fshift)/(fMax-1)));
      fthumb.right := fthumb.left + w;
      if fthumb.left < 0 then begin
        fthumb.left := 0;
        fthumb.right := fthumb.left + w;
      end;
      if fthumb.right > (width-fshift) then begin
        fthumb.right := (width-fshift);
        fthumb.left  := fthumb.right - w;
      end;
    end;
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSScrollBar.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
var
  w: integer;
begin
  inherited MouseDown(Button, Shift, X, Y);
  fmouse := Point(X,Y);
  fdown  := (Button = mbLeft);
  if PtInRect(fthumb,fmouse) then
    fbounds := fthumb
  else if fdown  then begin
    if fKind = sbVertical then begin
      w := (fthumb.bottom - fthumb.top) div 2;
      fthumb.top    := fmouse.y - w;
      fthumb.bottom := fmouse.y + w;
      SetPos(round(fthumb.top/(height/(fMax+1))));
    end else if X < width - fshift then begin
      w := (fthumb.right - fthumb.left) div 2;
      fthumb.left    := fmouse.x - w;
      fthumb.right   := fmouse.x + w;
      SetPos(round(fthumb.left/((width-fshift)/(fMax+1))));
    end;
    fbounds := fthumb;
    fdown   := true;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TSScrollBar.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if fdown then begin
    if (fKind = sbVertical) then begin
      fthumb.top    := fbounds.top    - (fmouse.y - Y);
      fthumb.bottom := fbounds.bottom - (fmouse.y - Y);
      SetPos(round(fthumb.top/((height-integer(odd(fMax)))/(fMax+1)))); //!!!!!
    end else begin
      fthumb.left    := fbounds.left  - (fmouse.x - X);
      fthumb.right   := fbounds.right - (fmouse.x - X);
      SetPos(round(fthumb.left/((width-fshift)/(fMax+1))));
    end;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TSScrollBar.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  fdown   := false;
  fbounds := fthumb;
  InvalidateRect(handle,nil,false);
end;

procedure TSScrollBar.Resize;
var
  w: integer;
begin
  if fKind = sbVertical then begin
    w := fthumb.bottom - fthumb.top;
    fthumb.top := round(fPos * height/(fMax+1));
    fthumb.bottom := fthumb.top + w;
  end else begin
    w := fthumb.right - fthumb.left;
    fthumb.left  := round(fPos * (width-fshift)/(fMax+1));
    fthumb.right := fthumb.left + w;
  end;
end;

procedure TSScrollBar.Paint;
begin
  inherited Paint;

  canvas.pen.color := themes.scroll.border;
  canvas.pen.width := 3;
  if fKind = sbVertical then begin
    canvas.line(0,0,0,height);
  end else begin
    canvas.line(0,0,(width-fshift),0);
    if fshift > 0 then begin
      canvas.pen.width := 1;
      canvas.line(width-fshift,0,width,height);
    end;
  end;

  if fshow then begin
    canvas.brush.style := bsSolid;
    canvas.brush.color := themes.scroll.border;
    canvas.Rectangle(fthumb);
  end;

end;

{ TSShellView }

constructor TSShellView.Create(AOwner: TComponent);
var
  png : TPortableNetworkGraphic;
  vStream : TResourceStream;
begin
  inherited Create(AOwner);

  fimages := TCustomImageList.Create(Self);
  fimages.width  := 48;
  fimages.height := 48;
  png:=TPortableNetworkGraphic.Create;
  try
    vStream := TResourceStream.Create(HINSTANCE,'shell',RT_RCDATA);
    try
      Png.LoadFromStream(vStream);
      fimages.Add(png,nil);
    finally
      vStream.Free;
    end;
  finally
    Png.Free;
  end;

  color := themes.dialogs.shell.color;
  font.size := themes.dialogs.shell.font.size;
  font.color := themes.dialogs.shell.font.color;

  fvscroll        := TSScrollBar.Create(self);
  fvscroll.Parent := self;
  fvscroll.align  := alRight;
  fvscroll.kind   := sbVertical;
  //fvscroll.OnChange := @ScrollChange;

  files := TStringList.Create;
  fdata := TStringList.Create;

  fselect := Rect(0,0,12+48+12,2+48+2+18*2);
  fheader := 25;

  fmultisel := false;
end;

destructor TSShellView.Destroy;
var
  i: integer;
begin
  for i:= 0 to files.count - 1 do begin
    files.objects[i].free;
  end;
  FreeAndNil(files);
  FreeAndNil(fdata);
  FreeAndNil(fimages);
  inherited Destroy;
end;

procedure TSShellView.ScanFile(str: string);
var
  i     : integer;
  SR,SRE: TSearchRec;
  shi   : TShellItem;
  path,s: string;
begin
  for i:= 0 to files.count - 1 do begin
    files.objects[i].free;
  end;
  files.Clear;
  fdata.Clear;
  path := str;
  if FindFirst(str+'\*',faAnyFile,SR) = 0 then
    repeat
      str:= SR.Name;
      if not ((str = '.')or(str = '..')or(str = '...'))then begin
        shi := TShellItem.Create;
        shi.attr := sr.Attr;
        shi.path := path;
        shi.ext  := ExtractFileExt(path+str);
        shi.selected := false;
        if sr.Attr = faDirectory then begin
          if FindFirst(path+str+'\*',faAnyFile,SRE) = 0 then begin
            i := 0;
            repeat
              s:= SRE.Name;
              if not ((s = '.')or(s = '..')or(s = '...'))then inc(i);
            until FindNext(SRE) <> 0;
            shi.isempty := i = 0;
            FindClose(SRE);
          end;
        end;
        files.AddObject(str,shi);
      end;
    until FindNext(SR) <> 0;
  FindClose(SR);
end;

procedure TSShellView.CheckSelected;
var
  i,x1,y1: integer;
  r: TRect;

  function OverlapRects(R1, R2: TRect): boolean;
  begin
    result := (R1.Left <= R2.Right) and (R2.Left <= R1.Right) and (R1.Top <= R2.Bottom) and (R2.Top <= R1.Bottom);
  end;

  function SetRect(x,y: TPoint): TRect;
  var
    z: TPoint;
  begin
    if x.x > y.x then begin z.x := x.x; x.x := y.x; y.x := z.x; end;
    if x.y > y.y then begin z.y := x.y; x.y := y.y; y.y := z.y; end;
    result := Rect(x.x,x.y,y.x,y.y);
  end;

begin
  x1 := 5;
  y1 := 5 + fheader;
  fdata.Clear;
  for i:= 0 to files.count - 1 do begin
    if (x1 + fselect.right + 5 > width - fvscroll.width) then begin
      x1 := 5;
      inc(y1,fselect.bottom+5);
    end;
    r.left   := x1;
    r.top    := y1;
    r.right  := x1+ fselect.right;
    r.bottom := y1+ fselect.bottom;
    TShellItem(files.objects[i]).selected := OverlapRects(r,SetRect(fstart,fend));
    if TShellItem(files.objects[i]).selected and (TShellItem(files.objects[i]).attr <> faDirectory) then begin
      fdata.Add(TShellItem(files.objects[i]).path+files[i]);
      if assigned(fonselect) then fonselect(self);
    end;
    inc(x1,fselect.right+5);
  end;
end;

procedure TSShellView.DblClick;
var
  i: integer;
begin
  inherited DblClick;
  i := 0;
  while i < files.count do begin
    if TShellItem(files.objects[i]).selected then begin
      if TShellItem(files.objects[i]).attr = faDirectory then begin
        ScanFile(TShellItem(files.objects[i]).path+files[i]+PathDelim);
        InvalidateRect(handle,nil,false);
        exit;
      end else begin
        fdata.Clear;
        fdata.Add(TShellItem(files.objects[i]).path+files[i]);
        if assigned(fondblclick) then fondblclick(self);
      end;
    end;
    inc(i);
  end;
end;

procedure TSShellView.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  fdown := (Button = mbLeft) and (Y > fheader);
  if fdown then begin
    fstart := Point(X,Y);
    fend   := fstart;
    CheckSelected();
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSShellView.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if fdown then begin
    fend := Point(X,Y);
    if fmultisel then CheckSelected();
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSShellView.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if fdown then begin
    fdown  := false;
    fend   := fstart;
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSShellView.Resize;
begin
  inherited Resize;

end;

procedure TSShellView.Paint;
var
  i,x,y,tw,tx,n,h: integer;
  str,ext: string;

  procedure Swap(x,y: TPoint);
  var
    z: TPoint;
  begin
    if x.x > y.x then begin z.x := x.x; x.x := y.x; y.x := z.x; end;
    if x.y > y.y then begin z.y := x.y; x.y := y.y; y.y := z.y; end;
    canvas.DrawFocusRect(Rect(x.x,x.y,y.x,y.y));
  end;

begin
  inherited Paint;

  canvas.pen.color := themes.scroll.border;
  canvas.pen.width := 3;
  canvas.Rectangle(0,0,width,height);

  x := 5;
  y := 5 + fheader;
  for i:= 0 to files.count - 1 do begin
    if (x + fselect.right + 5 > width - fvscroll.width) then begin
      x := 5;
      inc(y,fselect.bottom+5);
    end;
    if TShellItem(files.objects[i]).selected then begin
      canvas.pen.width := 1;
      canvas.brush.color := $784f26;
      canvas.brush.style := bsSolid;
      canvas.roundrect(x+fselect.left,y+fselect.top,x+fselect.right,y+fselect.bottom,5,5);
    end;
    if TShellItem(files.objects[i]).attr = faDirectory then begin
      if TShellItem(files.objects[i]).isempty then
        fimages.Draw(canvas,x+(fselect.right-fselect.left-fimages.width) div 2,y+2,4,true)
      else
        fimages.Draw(canvas,x+(fselect.right-fselect.left-fimages.width) div 2,y+2,6,true);
    end else begin
      n := 0;
      ext := lowercase(TShellItem(files.objects[i]).ext);
      if       (ext = '.jpg') then n := 1
      else if  (ext = '.gif')
      or       (ext = '.bmp')
      or       (ext = '.ico')
      or       (ext = '.png') then n := 2
      else if  (ext = '.mod') then n := 3
      else if  (ext = '.zip') then n := 7
      else if  (ext = '.rar') then n := 8;
      fimages.Draw(canvas,x+(fselect.right-fselect.left-fimages.width) div 2,y+2,n,true);
    end;

    canvas.brush.style := bsClear;
    tw := canvas.textwidth(files[i]);
    if 5+tw+5 < fselect.right then begin
      tx := (fselect.right-tw) div 2;
      canvas.textout(x+tx,y+2+fimages.height,files[i]);
    end else begin
      tx := 5;
      n := 1;
      h := 0;
      while (h < 2) and (n <= length(files[i])) do begin
        str := UTF8Copy(files[i],n,1);
        canvas.textout(x+tx,h*16+y+2+fimages.height,str);
        tx := tx + canvas.textwidth(str);
        inc(n);
        if 5+tx+5 > fselect.right then begin
          tx := 5;
          inc(h);
        end;
      end;
    end;
    inc(x,fselect.right+5);
  end;

  if fdown then Swap(fstart,fend);

  for i:= 0 to fdata.count - 1 do begin
    canvas.textout(5,height-30-i*20,fdata[i]);
  end;

end;

end.

