unit SDialogs;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Controls, Graphics, Forms, LCLIntf, LCLType, LMessages,
  SThemes, SForms, SControls;

type
  TMsgDlgType    = (mtWarning, mtError, mtInformation, mtConfirmation,mtCustom);
  TMsgDlgBtn     = (mbYes, mbNo, mbOK, mbCancel);
  TMsgDlgButtons =  set of TMsgDlgBtn;

const
  mbYesNoCancel      = [mbYes, mbNo, mbCancel];
  mbYesNo            = [mbYes, mbNo];
  mbOKCancel         = [mbOK, mbCancel];

type
  { TSDialog }

  TSDialog = class
  public
    procedure   Resize; virtual; abstract;
    procedure   Paint;  virtual; abstract;
  end;

  { TSDialogForm }

  TSDialogForm = class(TForm)
  protected
    {$IFDEF WINDOWS}procedure WndProc(var TheMessage: TLMessage); override;{$ENDIF}
    {$IFNDEF WINDOWS}
    procedure MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    {$ENDIF}
  private
    { private declarations }
    tw,th       : integer;
    fbitmap     : TBitmap;
    fclose      : TSysButton;
    floaded     : boolean;
    fbounds     : TRect;
    fbordersize : integer;
    fframe      : integer;
    {$IFNDEF WINDOWS}
    fmoving   : boolean;
    fsizing   : boolean;
    fdown   : boolean;
    fmouse  : TPoint;
    procedure ThisCursor(X,Y: Integer);
    {$ENDIF}

    procedure   SystemClick(Sender: TObject);
  public
    { public declarations }
    fparent    : TSDialog;
    constructor CreateNew(AOwner: TComponent; Num: Integer = 0); override;
    destructor  Destroy; override;
    procedure   Resize; override;
    procedure   Paint; override;
  end;

  { TSOpenDialog }

  TSOpenDialog = class(TSDialog)
  private
    { private declarations }
    fdialog   : TSDialogForm;
    fresult   : boolean;
    fcancel   : TSButton;
    fopen     : TSButton;
    fshell    : TSShellView;
    ffilename : string;
    ffiles    : TStringList;
    procedure   FilesDblClick(Sender: TObject);
    procedure   FilesSelect(Sender: TObject);
  public
    { public declarations }
    constructor Create;
    function    Execute(): boolean;
    procedure   Click(Sender: TObject);
    procedure   Resize; override;
    procedure   Paint;  override;
    property    Files: TStringList read ffiles;
    property    FileName: string read ffilename write ffilename;
  end;

  { TSSaveDialog }

  TSSaveDialog = class(TSOpenDialog)
  private
    { private declarations }
  public
    { public declarations }
    constructor Create;
    procedure   Click(Sender: TObject);
  end;

  { TSMessageBox }

  TSMessageBox = class(TSDialog)
  private
    { private declarations }
    fdialog : TSDialogForm;
    fresult : TMsgDlgBtn;
    fcancel : TSButton;
    fok     : TSButton;
    fno     : TSButton;
    fyes    : TSButton;
    fmsg    : string;
  public
    { public declarations }
    constructor Create;
    procedure   Click(Sender: TObject);
    procedure   Resize; override;
    procedure   Paint;  override;
  end;

  function MessageDlg(const aMsg: string; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons): TMsgDlgBtn;
  procedure ShowMessage(const msg: string);

implementation

{ TSDialogForm }

{$IFDEF WINDOWS}
procedure TSDialogForm.WndProc(var TheMessage: TLMessage);
var
  mx,my : longint;
  r     : TRect;
const
  HTBOTTOM = 15;
  HTBOTTOMLEFT = 16;
  HTBOTTOMRIGHT = 17;
  HTCAPTION = 2;
  HTLEFT = 10;
  HTRIGHT = 11;
  HTTOP = 12;
  HTTOPLEFT = 13;
  HTTOPRIGHT = 14;
begin
  case TheMessage.Msg of
    LM_NCHITTEST :
                  begin
                    mx := WORD(TheMessage.lParam);
                    my := WORD(((DWORD(TheMessage.lParam)) shr 16) and $FFFF);
                    r.left   := mx - boundsrect.left;
                    r.right  := boundsrect.right - mx;
                    r.top    := my - boundsrect.top;
                    r.bottom := boundsrect.bottom - my;
                    if (r.top < themes.form.border.size) and (r.left < themes.form.border.size) then
                      TheMessage.Result := HTTOPLEFT
                    else if (r.top < themes.form.border.size) and (r.right < themes.form.border.size) then
                      TheMessage.Result := HTTOPRIGHT
                    else if (r.bottom < themes.form.border.size) and (r.left < themes.form.border.size) then
                      TheMessage.Result := HTBOTTOMLEFT
                    else if (r.bottom < themes.form.border.size) and (r.right < themes.form.border.size) then
                      TheMessage.Result := HTBOTTOMRIGHT
                    else if (r.top < themes.form.border.size) then
                      TheMessage.Result := HTTOP
                    else if (r.left < themes.form.border.size) then
                      TheMessage.Result := HTLEFT
                    else if (r.bottom < themes.form.border.size) then
                      TheMessage.Result := HTBOTTOM
                    else if (r.right < themes.form.border.size) then
                      TheMessage.Result := HTRIGHT
                    else if (r.top > themes.form.border.size) and (r.top < themes.form.border.frame) then
                      TheMessage.Result := HTCAPTION
                    else
                      inherited WndProc(TheMessage);
                  end;
  else
    inherited WndProc(TheMessage);
  end;
end;
{$ENDIF}

{$IFNDEF WINDOWS}

procedure TSDialogForm.ThisCursor(X,Y: Integer);
begin
  if      PtInRect(Rect(0,themes.form.border.size+5,themes.form.border.size,height-themes.form.border.size-5),Point(X,Y)) then            cursor := crSizeE
  else if PtInRect(Rect(0,0,themes.form.border.size+5,themes.form.border.size+5),Point(X,Y)) then                                          cursor := crSizeSE
  else if PtInRect(Rect(themes.form.border.size+5,0,width-themes.form.border.size-5,themes.form.border.size),Point(X,Y)) then             cursor := crSizeN
  else if PtInRect(Rect(width-themes.form.border.size-5,0,width,themes.form.border.size+5),Point(X,Y)) then                                cursor := crSizeSW
  else if PtInRect(Rect(width-themes.form.border.size,themes.form.border.size+5,width,height-themes.form.border.size-5),Point(X,Y)) then  cursor := crSizeW
  else if PtInRect(Rect(width-themes.form.border.size-5,height-themes.form.border.size-5,width,height),Point(X,Y)) then                    cursor := crSizeNW
  else if PtInRect(Rect(themes.form.border.size+5,height-themes.form.border.size,width-themes.form.border.size-5,height),Point(X,Y)) then cursor := crSizeS
  else if PtInRect(Rect(0,height-themes.form.border.size-5,themes.form.border.size+5,height),Point(X,Y)) then                              cursor := crSizeNE
  else cursor := crDefault;
end;

procedure TSDialogForm.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
var
  i: integer;
begin
  inherited MouseDown(Button, Shift, X, Y);
  fdown   := true;
  GetCursorPos(fmouse);
  fbounds := Rect(left,top,left+width,top+height);
  fsizing := cursor <> crDefault;
  fmoving := not fsizing and PtInRect(Rect(themes.form.border.size,themes.form.border.size,width-themes.form.border.size,themes.form.border.size+themes.form.border.frame),Point(X,Y));
  for i:=0 to ControlCount-1 do Controls[i].Enabled:=false;
end;

procedure TSDialogForm.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  p      : TPoint;
  l,t,w,h: integer;
begin
  inherited MouseMove(Shift, X, Y);

  if not fmoving and not fsizing then ThisCursor(X,Y);

  if fdown then begin

    GetCursorPos(p);

    if fmoving then begin
      left := fbounds.left +  p.x - fmouse.x;
      top  := fbounds.top  +  p.y - fmouse.y;
    end;

    if fsizing then begin

      l := left;
      t := top;
      w := width;
      h := height;

      case cursor of
        crSizeE  :
                   begin
                     l := fbounds.left +  p.x - fmouse.x;
                     w := (fbounds.right-fbounds.left) - (p.x - fmouse.x);
                   end;
        crSizeN  :
                   begin
                     t := fbounds.top  +  p.y - fmouse.y;
                     h := (fbounds.bottom-fbounds.top) - (p.y - fmouse.y);
                   end;
        crSizeW  : w := (fbounds.right-fbounds.left) + (p.x - fmouse.x);
        crSizeS  : h := (fbounds.bottom-fbounds.top) + (p.y - fmouse.y);

        crSizeSE :
                   begin
                     l := fbounds.left +  p.x - fmouse.x;
                     w := (fbounds.right-fbounds.left) - (p.x - fmouse.x);
                     t := fbounds.top  +  p.y - fmouse.y;
                     h := (fbounds.bottom-fbounds.top) - (p.y - fmouse.y);
                   end;

        crSizeSW :
                   begin
                     w := (fbounds.right-fbounds.left) + (p.x - fmouse.x);
                     t := fbounds.top  +  p.y - fmouse.y;
                     h := (fbounds.bottom-fbounds.top) - (p.y - fmouse.y);
                   end;

        crSizeNW :
                   begin
                     w := (fbounds.right-fbounds.left) + (p.x - fmouse.x);
                     h := (fbounds.bottom-fbounds.top) + (p.y - fmouse.y);
                   end;

        crSizeNE :
                   begin
                     h := (fbounds.bottom-fbounds.top) + (p.y - fmouse.y);
                     l := fbounds.left +  p.x - fmouse.x;
                     w := (fbounds.right-fbounds.left) - (p.x - fmouse.x);
                   end;
      end;

      if w < constraints.minwidth then begin
        w := constraints.minwidth;
        l := left;
      end;
      if h < constraints.minheight then begin
        h := constraints.minheight;
        t := top;
      end;

      SetBounds(l,t,w,h);
      {$IFDEF WINDOWS}Update;{$ENDIF}
    end;

  end;

end;

procedure TSDialogForm.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i: integer;
begin
  inherited MouseUp(Button, Shift, X, Y);
  fdown   := false;
  fsizing := false;
  fmoving := false;
  ThisCursor(X,Y);
  for i:=0 to ControlCount-1 do Controls[i].Enabled:=true;
  Update;
end;
{$ENDIF}

constructor TSDialogForm.CreateNew(AOwner: TComponent; Num: Integer);
begin
  inherited CreateNew(AOwner, Num);
  BorderStyle := bsNone;
  fbordersize := themes.form.border.size;
  fframe      := themes.form.border.frame;
  fbitmap     := TBitmap.Create;
  fbitmap.SetSize(1,1);
  fbitmap.Canvas.Pixels[0,0] := themes.form.border.color;
  color       := themes.form.color;
  font.color  := themes.form.font.color;
  font.size   := themes.form.font.size;
  constraints.minwidth  := 300;
  constraints.minheight := 170;

  fclose          := TSysButton.Create(self);
  fclose.parent   := self;
  fclose.top      := themes.form.border.size;
  fclose.mode     := smClose;
  fclose.tag      := ord(smClose);
  fclose.OnClick  := @SystemClick;

  floaded         := false;
  fbounds         := Rect(left,top,left+width,top+height);
end;

destructor TSDialogForm.Destroy;
begin
  FreeAndNil(fbitmap);
  FreeAndNil(fclose);
  inherited Destroy;
end;

procedure TSDialogForm.SystemClick(Sender: TObject);
begin
  case TSysMode(TSysButton(Sender).tag) of
    smClose  : Close;
  end;
end;

procedure TSDialogForm.Resize;
begin
  if handleallocated then begin
    if floaded then begin
      fclose.left := width - fclose.width - themes.form.border.size;
    end else
      fbounds := BoundsRect;
    floaded := true;
    fparent.Resize;
  end;
end;

procedure TSDialogForm.Paint;
begin

  th := canvas.TextHeight(caption);
  tw := canvas.TextWidth(caption);

  canvas.StretchDraw(Rect(0,0,width,fbordersize),fbitmap);
  canvas.StretchDraw(Rect(0,fbordersize,fbordersize,height-fbordersize),fbitmap);
  canvas.StretchDraw(Rect(width-fbordersize,fbordersize,width,height-fbordersize),fbitmap);
  canvas.StretchDraw(Rect(0,height-fbordersize,width,height),fbitmap);

  canvas.StretchDraw(Rect(fbordersize,fbordersize,width-fbordersize,fbordersize+fframe),fbitmap);

  canvas.brush.style := bsClear;
  canvas.font.color  := themes.form.font.shadow;
  canvas.TextOut((width - tw) div 2 + 1,(fbordersize + fframe - th - 1) div 2 + 1,caption);
  canvas.font.color  := themes.form.font.color;
  canvas.TextOut((width - tw) div 2,(fbordersize + fframe - th - 1) div 2,caption);

  fparent.Paint;
end;

{ TSOpenDialog }

constructor TSOpenDialog.Create;
begin
  inherited Create;
  fdialog := TSDialogForm.CreateNew(Application);
  fdialog.Position := poMainFormCenter;
  fdialog.caption  := 'Open file';
  fdialog.fparent  := self;
  fdialog.fbitmap.Canvas.Pixels[0,0] := themes.dialogs.open.border;
  fdialog.fclose.color   := themes.dialogs.open.border;
  fdialog.fclose.bgcolor := themes.dialogs.open.border;

  fshell          := TSShellView.Create(fdialog);
  fshell.parent   := fdialog;
  fshell.left     := themes.form.border.size +5;
  fshell.top      := themes.form.border.size +themes.form.border.frame + 5;
  fshell.width    := fdialog.width - themes.form.border.size*2 -10;
  fshell.height   := fdialog.height - themes.form.border.frame - themes.form.border.size*2 - 5 - 60;
  fshell.Anchors  := [akLeft, akTop, akRight, akBottom];
  fshell.OnDblClick := @FilesDblClick;
  fshell.OnSelect   := @FilesSelect;
  fshell.MultiSelect:= true;

  fcancel         := TSButton.Create(fdialog);
  fcancel.parent  := fdialog;
  fcancel.left    := fdialog.width - fcancel.width - themes.form.border.size - 10;
  fcancel.top     := fdialog.height - fcancel.height - themes.form.border.size - 15;
  fcancel.Anchors := [akRight, akBottom];
  fcancel.caption := 'Cancel';
  fcancel.OnClick := @Click;
  fcancel.tag     := 0;

  fopen           := TSButton.Create(fdialog);
  fopen.parent    := fdialog;
  fopen.left      := fdialog.width - fcancel.width - themes.form.border.size - 10 - fopen.width - 5;
  fopen.top       := fdialog.height - fopen.height - themes.form.border.size - 15;
  fopen.Anchors   := [akRight, akBottom];
  fopen.caption   := 'Open';
  fopen.OnClick   := @Click;
  fopen.tag       := 1;

  fresult         := false;
end;

function TSOpenDialog.Execute(): boolean;
begin
  fresult := false;
  fdialog.width    := 655;
  fdialog.height   := 480;
  fshell.ScanFile(ExtractFilePath(Application.ExeName));
  ffiles    := nil;
  ffilename := '';
  fdialog.ShowModal;
  ffiles    := fshell.FileData;
  if (ffiles <> nil) and (ffiles.count > 0) then begin
    ffilename := ffiles[0];
  end else
    fresult := false;
  result := fresult;
end;

procedure TSOpenDialog.Click(Sender: TObject);
begin
  case TSButton(Sender).tag of
    0: fdialog.close;
    1: begin fresult := true; fdialog.close; end;
  end;
end;

procedure TSOpenDialog.FilesDblClick(Sender: TObject);
begin
  fresult := true;
  fdialog.close;
end;

procedure TSOpenDialog.FilesSelect(Sender: TObject);
begin

end;

procedure TSOpenDialog.Resize;
begin

end;

procedure TSOpenDialog.Paint;
begin
  fdialog.canvas.Draw(themes.form.border.size + 5,(themes.form.border.size + themes.form.border.frame - Application.Icon.Height) div 2, Application.Icon);
  fdialog.canvas.pen.color := themes.dialogs.panel;
  fdialog.canvas.brush.color := themes.dialogs.panel;
  fdialog.canvas.brush.style := bsSolid;
  fdialog.canvas.rectangle(themes.form.border.size,fdialog.height-60,fdialog.width-themes.form.border.size,fdialog.height-themes.form.border.size);
end;

{ TSSaveDialog }

constructor TSSaveDialog.Create;
begin
  inherited Create;
  fdialog.caption  := 'Save file';
  fdialog.fbitmap.Canvas.Pixels[0,0] := themes.dialogs.save.border;
  fdialog.fclose.color   := themes.dialogs.save.border;
  fdialog.fclose.bgcolor := themes.dialogs.save.border;
  fopen.caption   := 'Save';
  fopen.OnClick   := @Click;
  fshell.MultiSelect:= false;
end;

procedure TSSaveDialog.Click(Sender: TObject);
begin
  case TSButton(Sender).tag of
    0: fresult := false;
    1: fresult := true;
  end;
  fdialog.close;
end;

{ TSMessageBox }

var
  fmessagebox: TSMessageBox = nil;

constructor TSMessageBox.Create;
begin
  inherited Create;
  fdialog := TSDialogForm.CreateNew(Application);

  fdialog.fbordersize := themes.form.border.size - 3;
  fdialog.fframe      := themes.form.border.frame + 3;

  fdialog.Position := poMainFormCenter;
  fdialog.width    := 355;
  fdialog.height   := 170;
  fdialog.constraints.minwidth  := fdialog.width;
  fdialog.constraints.minheight := fdialog.height;
  fdialog.constraints.maxwidth  := fdialog.constraints.minwidth;
  fdialog.constraints.maxheight := fdialog.constraints.minheight;
  fdialog.caption  := '';
  fdialog.fparent  := self;

  fcancel         := TSButton.Create(fdialog);
  fcancel.parent  := fdialog;
  fcancel.visible := false;
  fcancel.Anchors := [akRight, akBottom];
  fcancel.caption := 'Cancel';
  fcancel.OnClick := @Click;
  fcancel.tag     := 0;

  fok             := TSButton.Create(fdialog);
  fok.parent      := fdialog;
  fok.visible     := false;
  fok.Anchors     := [akRight, akBottom];
  fok.caption     := 'OK';
  fok.OnClick     := @Click;
  fok.tag         := 1;

  fno             := TSButton.Create(fdialog);
  fno.parent      := fdialog;
  fno.visible     := false;
  fno.Anchors     := [akRight, akBottom];
  fno.caption     := 'NO';
  fno.OnClick     := @Click;
  fno.tag         := 2;

  fyes            := TSButton.Create(fdialog);
  fyes.parent     := fdialog;
  fyes.visible    := false;
  fyes.Anchors    := [akRight, akBottom];
  fyes.caption    := 'YES';
  fyes.OnClick    := @Click;
  fyes.tag        := 3;

  fresult         := mbCancel;
end;

procedure TSMessageBox.Click(Sender: TObject);
begin
  case TSButton(Sender).tag of
    0: fresult := mbCancel;
    1: fresult := mbOK;
    2: fresult := mbNo;
    3: fresult := mbYes;
  end;
  fdialog.close;
end;

procedure TSMessageBox.Resize;
begin

end;

procedure TSMessageBox.Paint;
begin
  fdialog.canvas.brush.color := themes.dialogs.panel;
  fdialog.canvas.pen.color   := themes.dialogs.panel;
  fdialog.canvas.rectangle(fdialog.fbordersize,fdialog.height-40,fdialog.width-fdialog.fbordersize,fdialog.height-fdialog.fbordersize);
  fdialog.canvas.brush.style := bsClear;
  fdialog.canvas.textout(fdialog.fbordersize+10,fdialog.fframe+20,fmsg);
end;

function MessageDlg(const aMsg: string; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons): TMsgDlgBtn;
var
  clr: TColor;
  l,t: integer;
begin
  if fmessagebox = nil then fmessagebox:= TSMessageBox.Create;
  fmessagebox.fresult := mbCancel;
  clr := $333333;
  case DlgType of
    mtWarning      : begin clr := $024fae; fmessagebox.fdialog.caption := 'Warning!'; end;
    mtError        : begin clr := $000080; fmessagebox.fdialog.caption := 'Error!'; end;
    mtInformation  : begin clr := $925b02; fmessagebox.fdialog.caption := 'Info!'; end;
    mtConfirmation : begin clr := $008000; fmessagebox.fdialog.caption := 'Confirmation!'; end;
    mtCustom       : begin clr := $333333; fmessagebox.fdialog.caption := 'Message!'; end;
  end;
  fmessagebox.fdialog.fbitmap.Canvas.Pixels[0,0] := clr;
  fmessagebox.fdialog.fclose.color   := clr;
  fmessagebox.fdialog.fclose.bgcolor := clr;

  fmessagebox.fmsg := aMsg;

  fmessagebox.fcancel.visible := false;
  fmessagebox.fok.visible     := false;
  fmessagebox.fno.visible     := false;
  fmessagebox.fyes.visible    := false;

  l := fmessagebox.fdialog.width  - fmessagebox.fdialog.fbordersize;
  t := fmessagebox.fdialog.height - fmessagebox.fcancel.height - fmessagebox.fdialog.fbordersize - 5;
  if mbCancel in Buttons then begin
    l := l - fmessagebox.fcancel.width - 5;
    fmessagebox.fcancel.left := l;
    fmessagebox.fcancel.top  := t;
    fmessagebox.fcancel.visible := true;
  end;
  if mbOk in Buttons then begin
    l := l - fmessagebox.fok.width - 5;
    fmessagebox.fok.left := l;
    fmessagebox.fok.top  := t;
    fmessagebox.fok.visible := true;
  end;
  if mbNo in Buttons then begin
    l := l - fmessagebox.fno.width - 5;
    fmessagebox.fno.left := l;
    fmessagebox.fno.top  := t;
    fmessagebox.fno.visible := true;
  end;
  if mbYes in Buttons then begin
    l := l - fmessagebox.fyes.width - 5;
    fmessagebox.fyes.left := l;
    fmessagebox.fyes.top  := t;
    fmessagebox.fyes.visible := true;
  end;

  fmessagebox.fdialog.ShowModal;
  result := fmessagebox.fresult;

end;

procedure ShowMessage(const msg: string);
begin
  MessageDlg(msg,mtCustom,[mbOk]);
end;

finalization
  if assigned(fmessagebox) then fmessagebox.free;

end.

