unit SForms;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, SThemes,
  LCLIntf, LCLType, Graphics, ExtCtrls, LMessages;

type
  TMenuEvent = procedure(Sender: TObject; btnindex: integer; menuindex: integer) of object;
  TSysMode = (smClose,smReturn,smMinimal,smToggle);
  TFormStatus = (fsMaximized,fsNormal);
  TSysPanelMode = (spLeft,spTop);
  TSystemPanel = class;
  TPanelButton = class;
  TDataObj = class
    imageindex: integer;
  end;

  { TSysButton }

  TSysButton = class(TCustomControl)
  protected
    procedure MouseEnter; override;
    procedure MouseLeave; override;
    procedure MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
  private
    { private declarations }
    fmode    : TSysMode;
    fstat    : boolean;
    fonclick : TNotifyEvent;
  public
    { public declarations }
    bgcolor     : TColor;
    constructor Create(TheOwner: TComponent);override;
    destructor  Destroy; override;
    procedure   Paint; override;
    property    Mode  : TSysMode read fmode write fmode;
    property    Status: boolean read fstat;
    property    OnClick: TNotifyEvent read fonclick write fonclick;
  end;

  {TSPopup}

  TSPopup = class(TForm)
  protected
    procedure WMKillFocus(var Message: TLMKillFocus); message LM_KILLFOCUS;
    procedure MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseLeave; override;
  private
    { private declarations }
    fnum       : integer;
    fpos       : integer;
    fstrings   : TStringList;
    fimage     : TImageList;
    fmenunum   : integer;
    fdown      : boolean;
    fparent    : TPanelButton;
    fpanel     : TSystemPanel;
  public
    { public declarations }
    constructor CreateNew(AOwner: TComponent; Num: Integer = 0); override;
    destructor  Destroy; override;
    procedure   DoShow; override;
    procedure   Paint; override;
  end;

  { TPanelButton }

  TPanelButton = class(TCustomControl)
  protected
    procedure MouseEnter; override;
    procedure MouseLeave; override;
    procedure MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
  private
    { private declarations }
    fparent     : TSystemPanel;
    fnum        : integer;
    ftime       : integer;
    ftimer      : TTimer;
    fpopup      : TSPopup;
    fmenu       : TStringList;
    fonclick    : TMenuEvent;
    fonmenuclick: TMenuEvent;
    fdeficon    : integer;
    fimagesmall : TImageList;
    fimagebig   : TImageList;
    fseparr     : array [0..50] of boolean;
    procedure   TimerTimer(Sender: TObject);
    procedure   ClickMenu(index: integer);
    procedure   ShowPopup;
    procedure   HidePopup;
  public
    { public declarations }
    constructor Create(TheOwner: TSystemPanel); overload;
    destructor  Destroy; override;
    procedure   Paint; override;
    procedure   AddMenu(txt: string; index: integer = -1);
    property    DefaultIcon: integer read fdeficon write fdeficon;
    property    ImageBig: TImageList read fimagebig write fimagebig;
    property    ImageSmall: TImageList read fimagesmall write fimagesmall;
    property    OnClick: TMenuEvent read FOnClick write FOnClick;
    property    OnMenuClick: TMenuEvent read FOnMenuClick write FOnMenuClick;
  end;

  { TSystemPanel }

  TSystemPanel = class(TCustomControl)
  protected
  private
    { private declarations }
    fframe      : TCustomControl;
    fwidth      : integer;
    fmode       : TSysPanelMode;
    fbuttons    : TStringList;
    fswitch     : boolean;
    procedure   SetMode(AValue: TSysPanelMode);
    procedure   FrameClick(Sender: TObject);
  public
    { public declarations }
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    function    CanFocus: Boolean; override;
    procedure   ShowFrame(exec: boolean);
    function    AddButton(txt: string): TPanelButton;
    procedure   Resize; override;
    procedure   Paint; override;
    property    Mode: TSysPanelMode read fmode write SetMode;
  end;

  { TSForm }

  TSForm = class(TForm)
  protected
    {$IFDEF WINDOWS}procedure WndProc(var TheMessage: TLMessage); override;{$ENDIF}
    {$IFNDEF WINDOWS}
    procedure MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    {$ENDIF}
    procedure MouseLeave; override;
  private
    { private declarations }
    tw,th     : integer;
    fbitmap   : TBitmap;
    fbounds   : TRect;
    fmoving   : boolean;
    fsizing   : boolean;
    fpanel    : TSystemPanel;
    fclose    : TSysButton;
    freturn   : TSysButton;
    fminimal  : TSysButton;
    ftoggle   : TSysButton;
    floaded   : boolean;
    fstatus   : TFormStatus;
    ftray     : TTrayIcon;
    {$IFNDEF WINDOWS}
    fdown   : boolean;
    fmouse  : TPoint;
    procedure ThisCursor(X,Y: Integer);
    {$ENDIF}
    procedure SystemClick(Sender: TObject);
    procedure TrayClick(Sender: TObject);
  public
    { public declarations }
    sbheight : integer;
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Paint; override;
    procedure   Resize; override;
    property    SysPanel : TSystemPanel read fpanel write fpanel;
  end;

var
  Screen : Forms.TScreen;

implementation

{ TSysButton }

constructor TSysButton.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  width := 31;
  height := 25;
  color := themes.buttons.color;
  bgcolor := themes.buttons.color;
  fmode := smClose;
  fstat := false;
end;

destructor TSysButton.Destroy;
begin

  inherited Destroy;
end;

procedure TSysButton.MouseEnter;
begin
  inherited MouseEnter;
  color := $0000CC;
  if fmode <> smClose then color := themes.buttons.enter;
end;

procedure TSysButton.MouseLeave;
begin
  inherited MouseLeave;
  color := bgcolor;
end;

procedure TSysButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  color := $7a70f1;
  if fmode <> smClose then color := themes.buttons.light;
end;

procedure TSysButton.MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  color := clRed;
  if fmode <> smClose then color := themes.buttons.light;
  if PtInRect(ClientRect,Point(X,Y)) then fstat := not fstat;
  inherited MouseUp(Button, Shift, X, Y);
  if PtInRect(ClientRect,Point(X,Y)) then begin
    if assigned(fonclick) then fonclick(self);
  end else
    color := themes.buttons.color;
end;

procedure TSysButton.Paint;
var
  cx,cy: integer;
begin
  cx := width div 2;
  cy := height div 2;
  canvas.pen.color := $ffffff;
  canvas.Pen.width := 1;
  case fmode of
    smClose  :
              begin
                canvas.Line(cx-4,cy-4,cx+5,cy+5);
                canvas.Line(cx+4,cy-4,cx-5,cy+5);
                canvas.Line(cx,cy-1,cx,cy+2);
              end;
    smReturn :
                if TSForm(parent).fstatus = fsMaximized then begin
                  canvas.Rectangle(cx-4,cy-6,cx+6,cy+4);
                  canvas.Rectangle(cx-6,cy-4,cx+4,cy+6);
                end else
                  canvas.Rectangle(cx-5,cy-5,cx+5,cy+5);
    smMinimal:  canvas.Line(cx-5,cy,cx+5,cy);
    smToggle :
              begin
                if not fstat then begin
                  canvas.Line(cx-4,cy,cx+1,cy-5);
                  canvas.Line(cx-4,cy,cx+1,cy+5);
                  canvas.Line(cx,cy,cx+5,cy-5);
                  canvas.Line(cx,cy,cx+5,cy+5);
                end else begin
                  canvas.Line(cx-5,cy-3,cx+5,cy-3);
                  canvas.Line(cx-5,cy,cx+5,cy);
                  canvas.Line(cx-5,cy+3,cx+5,cy+3);
                end;
              end;
  end;
end;

{ TSPopup }

constructor TSPopup.CreateNew(AOwner: TComponent; Num: Integer);
begin
  inherited CreateNew(AOwner, Num);
  fparent := TPanelButton(AOwner);
  BorderStyle := bsNone;
  color := themes.form.border.color;
  formstyle := fsStayOnTop;
  width     := 10+32+32*2;
  height    := 34;
  font.size := themes.form.font.size;
  fstrings  := nil;
  fdown     := false;
  fimage    := nil;
  fnum      := -1;
  fpos      := 7;
  ShowInTaskBar := stNever;
end;

destructor TSPopup.Destroy;
begin
  inherited Destroy;
end;

procedure TSPopup.DoShow;
var
  bmp: TBitmap;
  pts: array of TPoint;
begin
  inherited DoShow;

  //SetWindowLong(Handle, GWL_STYLE,GetWindowLong(Handle, GWL_STYLE) and not WS_BORDER and not WS_OVERLAPPEDWINDOW);
  //SetWindowLong(Handle, GWL_EXSTYLE,GetWindowLong(Handle, GWL_EXSTYLE) or WS_EX_TOOLWINDOW and not WS_EX_APPWINDOW);

  bmp := TBitmap.Create;
  bmp.Monochrome := True;
  bmp.Width := Width;
  bmp.Height := Height;

  if fstrings.count>0 then begin
    case fpanel.fmode of
      spLeft :
                begin
                  SetLength(pts, 7);
                  pts[0] := Point(0,0);
                  pts[1] := Point(10,0);
                  pts[2] := Point(10,fpos);
                  pts[3] := Point(0,fpos+10);
                  pts[4] := Point(10,fpos+20);
                  pts[5] := Point(10,Height);
                  pts[6] := Point(0,Height);
                end;
      spTop  :
                begin
                  SetLength(pts, 7);
                  pts[0] := Point(0,0);
                  pts[1] := Point(0,10);
                  pts[2] := Point(width div 2 - 10,10);
                  pts[3] := Point(width div 2,0);
                  pts[4] := Point(width div 2 + 10,10);
                  pts[5] := Point(width,10);
                  pts[6] := Point(width,0);
                end;
    end;
  end else begin
    SetLength(pts, 4);
    pts[0] := Point(0,0);
    pts[1] := Point(width,0);
    pts[2] := Point(width,height);
    pts[3] := Point(0,height);
  end;

  with bmp.Canvas do begin
    Brush.Color := clWhite;
    FillRect(0,0,bmp.Width, bmp.Height);
    Brush.Color := clBlack;
    Polygon(pts);
  end;
  SetShape(bmp);
  bmp.Free;

  fmenunum := 0;
end;

procedure TSPopup.WMKillFocus(var Message: TLMKillFocus);
begin
  fmenunum := 0;
  fparent.color := themes.form.border.color;
  InvalidateRect(fparent.handle,nil,false);
  Hide;
  Message.Result := 1;
end;

procedure TSPopup.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  fdown := true;
  InvalidateRect(handle,nil,false);
end;

procedure TSPopup.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  oh: integer;
begin
  inherited MouseMove(Shift, X, Y);
  if not fdown then begin
    case fpanel.fmode of
      spLeft : oh := 0;
      spTop  : oh := 10;
    end;
    fmenunum := (Y-oh) div 32 + 1;
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSPopup.MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  fdown := false;
  fpanel.fswitch := false;
  if (fmenunum = Y div 32 + 1) then begin
    fnum           := fmenunum - 1;
    fparent.ClickMenu(fnum);
    Hide;
    exit;
  end else
    fmenunum := Y div 32 + 1;
  InvalidateRect(handle,nil,false);
end;

procedure TSPopup.MouseLeave;
begin
  fmenunum := 0;
  InvalidateRect(handle,nil,false);
  inherited MouseLeave;
end;

procedure TSPopup.Paint;
var
  i,idx,th,ow,oh: integer;
begin

  case fpanel.fmode of
    spLeft :
             begin
               ow := 10;
               oh := 0;
             end;
    spTop  :
             begin
               ow := 0;
               oh := 10;
             end;
  end;

  if fdown then begin
    canvas.brush.color := themes.buttons.hight;
    canvas.pen.color := themes.buttons.hight;
  end else begin
    canvas.brush.color := themes.buttons.light;
    canvas.pen.color := themes.buttons.light;
  end;
  if fmenunum > 0 then
    canvas.Rectangle(ow+1,oh+(fmenunum-1)*32,width,oh+(fmenunum-1)*32+32);

  i := 0;
  while i<height do begin
    canvas.pixels[ow+32,i] := clWhite;
    inc(i,2);
  end;

  if fstrings<>nil then begin
    canvas.brush.style := bsClear;
    th := canvas.textheight('X');
    for i:=0 to fstrings.count-1 do begin

      idx := TDataObj(fstrings.objects[i]).imageindex;
      if (fparent.fimagesmall<>nil) and (idx <> -1) then
        fparent.fimagesmall.Draw(canvas,ow + (32 - fparent.fimagesmall.width) div 2 ,oh+i*32 + (32 - fparent.fimagesmall.height) div 2 ,idx,true);

      canvas.font.color := themes.buttons.hight;
      canvas.textout(ow+32+5+1,oh+i*32 + (32 - th) div 2,fstrings[i]);
      canvas.font.color := $FFFFFF;
      canvas.textout(ow+32+5,oh+i*32 + (32 - th) div 2 - 1,fstrings[i]);

      if fparent.fseparr[i] then begin
        canvas.pen.color := themes.buttons.hight;
        canvas.Line(ow+32+1,oh+i*32,width,oh+i*32);
      end;

    end;
  end;

  canvas.brush.style := bsClear;
  canvas.pen.color   := themes.buttons.hight;
  case fpanel.fmode of
    spLeft :  canvas.Polygon([
                Point(11,0),Point(width-1,0),Point(width-1,height-1),Point(11,height-1),
                Point(11,fpos+20),Point(1,fpos+10),Point(11,fpos)
              ]);
    spTop  :  canvas.Polygon([
                Point(0,11),Point(width div 2 - 10,11),Point(width div 2,1),Point(width div 2 + 10,11), Point(width -1,11),
                Point(width -1,height-1),Point(0,height-1)
              ]);
  end;

end;

{ TPanelButton }

constructor TPanelButton.Create(TheOwner: TSystemPanel);
var
  i: integer;
begin
  inherited Create(TheOwner);
  fparent := TheOwner;
  width   := 48;
  height  := 48;
  color   := themes.form.border.color;
  ftimer  := TTimer.Create(self);
  ftimer.interval:= 10;
  ftimer.enabled := false;
  ftimer.ontimer := @TimerTimer;
  ftime          := 0;
  fmenu          := TStringList.Create;
  fpopup         := TSPopup.CreateNew(self);
  fpopup.fpanel  := TheOwner;
  fpopup.fstrings:= fmenu;
  fimagesmall    := nil;
  fimagebig      := nil;
  for i:=0 to length(fseparr)-1 do fseparr[i] := false;
end;

destructor TPanelButton.Destroy;
var
  i: integer;
begin
  ftimer.enabled := false;
  FreeAndNil(ftimer);
  FreeAndNil(fpopup);
  for i:=0 to fmenu.count - 1 do fmenu.Objects[i].free;
  FreeAndNil(fmenu);
  inherited Destroy;
end;

procedure TPanelButton.ShowPopup;
var
  p: TPoint;
begin
  p := ClientToScreen(Point(0,0));

  case fparent.fmode of
    spLeft :
             begin
               fpopup.height := 32*fmenu.Count;
               fpopup.left := p.x + width + 2;
               fpopup.top  := p.y;
             end;
    spTop  :
             begin
               fpopup.height := 10+32*fmenu.Count;
               fpopup.left := p.x + (width - fpopup.width) div 2;
               fpopup.top  := p.y + height + 4;
             end;
  end;
  if fpopup.height < 1 then fpopup.height := 1;
  fpopup.Show;
  fpopup.SetFocus;
end;

procedure TPanelButton.HidePopup;
var
  i: integer;
begin
  for i:=0 to fparent.ControlCount-1 do
    if (fparent.Controls[i] is TPanelButton) and (fparent.Controls[i] <> self) then
      TPanelButton(fparent.Controls[i]).fpopup.Hide;
end;

procedure TPanelButton.TimerTimer(Sender: TObject);
begin
  inc(ftime,10);
  fparent.fswitch := false;
  if ftime >= 500 then begin
    ftime := 0;
    ftimer.enabled := false;
    fparent.fswitch := true;
    ShowPopup;
  end;
end;

procedure TPanelButton.ClickMenu(index: integer);
begin
  if assigned(FOnMenuClick) then FOnMenuClick(self,fnum,index);
end;

procedure TPanelButton.MouseEnter;
begin
  inherited MouseEnter;
  color := themes.buttons.enter;
  if fparent.fswitch then begin
    HidePopup;
    ShowPopup;
  end;
end;

procedure TPanelButton.MouseLeave;
begin
  inherited MouseLeave;
  color := themes.form.border.color;
end;

procedure TPanelButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  color := themes.buttons.light;
  ftime := 0;
  if (Button=mbLeft) and (fmenu.Count>0) then ftimer.enabled := true;
end;

procedure TPanelButton.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if not PtInRect(BoundsRect,Point(X,Y)) then ftimer.enabled := false;
end;

procedure TPanelButton.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  color := themes.buttons.enter;
  ftime := 0;
  ftimer.enabled := false;
  if not fpopup.showing and PtInRect(ClientRect,Point(X,Y)) then begin
    if Button=mbLeft then begin
      if assigned(fonclick) then fonclick(self,fnum,fpopup.fnum);
    end else begin
      fpopup.fnum := -1;
    end;
    fparent.fswitch := false;
  end;
end;

procedure TPanelButton.AddMenu(txt: string; index: integer = -1);
var
  obj: TDataObj;
begin
  if txt = '-' then
    fseparr[fmenu.Count] := true
  else begin
    if fpopup.canvas.textwidth(txt)+10+32+5+10 > fpopup.width then
      fpopup.width := 10+32+5+fpopup.canvas.textwidth(txt)+10;
    fpopup.height := 32*(fmenu.Count+1);
    obj := TDataObj.Create;
    obj.imageindex := index;
    fmenu.AddObject(txt,obj);
  end;
end;

procedure TPanelButton.Paint;
begin
  if fpopup.showing then color := themes.buttons.light;
  if fmenu.count > 0 then begin
    canvas.pen.color := $FFFFFF;
    canvas.brush.color := $FFFFFF;
    if fparent.fmode = spLeft then
      canvas.Polygon([Point(width-5-4,4),Point(width-4,4),Point(width-4,4+5)])
    else
      canvas.Polygon([Point(width-5-4,height-4),Point(width-4,height-5-4),Point(width-4,height-4)]);
  end;
  if fimagebig <> nil then begin
    if fmenu.count = 0 then
      fimagebig.Draw(canvas,(width - fimagebig.width) div 2,(height - fimagebig.height) div 2,fdeficon,true)
    else begin
      if (fpopup.fnum = -1) or (fpopup.fnum > fimagebig.count - 1) then
        fimagebig.Draw(canvas,(width - fimagebig.width) div 2,(height - fimagebig.height) div 2,fdeficon,true)
      else
        fimagebig.Draw(canvas,(width - fimagebig.width) div 2,(height - fimagebig.height) div 2,fpopup.fnum,true);
    end;
  end;
end;

{ TSystemPanel }

constructor TSystemPanel.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  fwidth  := 54;
  fframe := TCustomControl.Create(self);
  fframe.parent := self;
  fframe.width  := fwidth;
  fframe.height := fwidth;
  fframe.align  := alLeft;
  fframe.color  := themes.form.border.color;
  fframe.OnClick:= @FrameClick;
  fmode         := spLeft;
  fbuttons      := TStringList.Create;
  fswitch       := false;
end;

destructor TSystemPanel.Destroy;
begin
  FreeAndNil(fframe);
  FreeAndNil(fbuttons);
  inherited Destroy;
end;

procedure TSystemPanel.FrameClick(Sender: TObject);
begin
  fswitch := false;
end;

function TSystemPanel.CanFocus: Boolean;
begin
  Result:=inherited CanFocus;
  fswitch := false;
end;

procedure TSystemPanel.SetMode(AValue: TSysPanelMode);
begin
  if fmode=AValue then Exit;
  fmode := AValue;
  case fmode of
    spLeft :
             begin
               if TSForm(parent).ftoggle.status then fframe.width := 0 else fframe.width := fwidth;
               fframe.align := alLeft;
             end;
    spTop  :
             begin
               if TSForm(parent).ftoggle.status then fframe.height := 0 else fframe.height := fwidth;
               fframe.align  := alTop;
             end;
  end;
end;

procedure TSystemPanel.ShowFrame(exec: boolean);
begin
  case fmode of
    spLeft : if exec then fframe.width  := 0 else fframe.width := fwidth;
    spTop  : if exec then fframe.height := 0 else fframe.height := fwidth;
  end;
end;

function TSystemPanel.AddButton(txt: string): TPanelButton;
begin
  result := TPanelButton.Create(self);
  result.fnum := fbuttons.count;
  fbuttons.AddObject(txt,result);
end;

procedure TSystemPanel.Resize;
var
  i: integer;
  b: TPanelButton;
begin
  inherited Resize;
  if parent <> nil then begin
    for i := 0 to fbuttons.count - 1 do begin
      b := TPanelButton(fbuttons.Objects[i]);
      b.parent := fframe;
      if fmode = spLeft then begin
        b.left   := 2;
        b.top    := i*b.height + 2;
        b.visible := b.top + b.height + 2 < height;
      end else begin
        b.left   := i*b.width + 2;
        b.top    := 2;
        b.visible := b.left + b.width + 2 < width;
      end;
    end;
  end;
end;

procedure TSystemPanel.Paint;
begin

end;

{ TSForm }

constructor TSForm.Create(TheOwner: TComponent);
begin
  inherited Create(Application);
  BorderStyle := bsNone;
  fbitmap     := TBitmap.Create;

  fbitmap.SetSize(1,1);
  fbitmap.Canvas.Pixels[0,0] := themes.form.border.color;
  color       := themes.form.color;
  font.color  := themes.form.font.color;
  font.size   := themes.form.font.size;

  constraints.minwidth  := 200;
  constraints.minheight := 170;

  th := canvas.TextHeight(caption);
  tw := canvas.TextWidth(caption);

  fbounds := Rect(left,top,left+width,top+height);
  fsizing := false;
  fmoving := false;

  fpanel          := TSystemPanel.Create(self);
  fpanel.parent   := self;
  fpanel.color    := themes.form.color;

  fclose          := TSysButton.Create(self);
  fclose.parent   := self;
  fclose.top      := themes.form.border.size;
  fclose.mode     := smClose;
  fclose.tag      := ord(smClose);
  fclose.OnClick  := @SystemClick;

  freturn         := TSysButton.Create(self);
  freturn.parent  := self;
  freturn.top     := themes.form.border.size;
  freturn.mode    := smReturn;
  freturn.tag     := ord(smReturn);
  freturn.OnClick := @SystemClick;

  fminimal        := TSysButton.Create(self);
  fminimal.parent := self;
  fminimal.top    := themes.form.border.size;
  fminimal.mode   := smMinimal;
  fminimal.tag    := ord(smMinimal);
  fminimal.OnClick:= @SystemClick;

  ftoggle         := TSysButton.Create(self);
  ftoggle.parent  := self;
  ftoggle.top     := themes.form.border.size;
  ftoggle.left    := themes.form.border.size + 5 + Application.Icon.width + 5;
  ftoggle.mode    := smToggle;
  ftoggle.tag     := ord(smToggle);
  ftoggle.OnClick := @SystemClick;

  floaded         := false;
  fstatus         := fsNormal;

  ftray           := TTrayIcon.Create(self);
  ftray.icon      := Application.icon;
  ftray.Hint      := caption;
  ftray.OnClick   := @TrayClick;

  sbheight        := 0;
end;

destructor TSForm.Destroy;
begin
  FreeAndNil(ftray);
  FreeAndNil(fpanel);
  FreeAndNil(fbitmap);
  FreeAndNil(fclose);
  FreeAndNil(freturn);
  FreeAndNil(fminimal);
  FreeAndNil(ftoggle);
  inherited Destroy;
end;

{$IFDEF WINDOWS}
procedure TSForm.WndProc(var TheMessage: TLMessage);
var
  mx,my : longint;
  r     : TRect;
const
  HTBOTTOM = 15;
  HTBOTTOMLEFT = 16;
  HTBOTTOMRIGHT = 17;
  HTCAPTION = 2;
  HTLEFT = 10;
  HTRIGHT = 11;
  HTTOP = 12;
  HTTOPLEFT = 13;
  HTTOPRIGHT = 14;
begin
  case TheMessage.Msg of
    LM_NCHITTEST :
                  begin
                    mx := WORD(TheMessage.lParam);
                    my := WORD(((DWORD(TheMessage.lParam)) shr 16) and $FFFF);
                    r.left   := mx - boundsrect.left;
                    r.right  := boundsrect.right - mx;
                    r.top    := my - boundsrect.top;
                    r.bottom := boundsrect.bottom - my;
                    if (r.top < themes.form.border.size) and (r.left < themes.form.border.size) then
                      TheMessage.Result := HTTOPLEFT
                    else if (r.top < themes.form.border.size) and (r.right < themes.form.border.size) then
                      TheMessage.Result := HTTOPRIGHT
                    else if (r.bottom < themes.form.border.size) and (r.left < themes.form.border.size) then
                      TheMessage.Result := HTBOTTOMLEFT
                    else if (r.bottom < themes.form.border.size) and (r.right < themes.form.border.size) then
                      TheMessage.Result := HTBOTTOMRIGHT
                    else if (r.top < themes.form.border.size) then
                      TheMessage.Result := HTTOP
                    else if (r.left < themes.form.border.size) then
                      TheMessage.Result := HTLEFT
                    else if (r.bottom < themes.form.border.size) then
                      TheMessage.Result := HTBOTTOM
                    else if (r.right < themes.form.border.size) then
                      TheMessage.Result := HTRIGHT
                    else if (r.top > themes.form.border.size) and (r.top < themes.form.border.frame) then
                      TheMessage.Result := HTCAPTION
                    else
                      inherited WndProc(TheMessage);
                  end;
  else
    inherited WndProc(TheMessage);
  end;
end;
{$ENDIF}

procedure TSForm.MouseLeave;
begin
  inherited MouseLeave;
  cursor := crDefault;
end;

{$IFNDEF WINDOWS}

procedure TSForm.ThisCursor(X,Y: Integer);
begin
  if      PtInRect(Rect(0,themes.form.border.size+5,themes.form.border.size,height-themes.form.border.size-5),Point(X,Y)) then            cursor := crSizeE
  else if PtInRect(Rect(0,0,themes.form.border.size+5,themes.form.border.size+5),Point(X,Y)) then                                          cursor := crSizeSE
  else if PtInRect(Rect(themes.form.border.size+5,0,width-themes.form.border.size-5,themes.form.border.size),Point(X,Y)) then             cursor := crSizeN
  else if PtInRect(Rect(width-themes.form.border.size-5,0,width,themes.form.border.size+5),Point(X,Y)) then                                cursor := crSizeSW
  else if PtInRect(Rect(width-themes.form.border.size,themes.form.border.size+5,width,height-themes.form.border.size-5),Point(X,Y)) then  cursor := crSizeW
  else if PtInRect(Rect(width-themes.form.border.size-5,height-themes.form.border.size-5,width,height),Point(X,Y)) then                    cursor := crSizeNW
  else if PtInRect(Rect(themes.form.border.size+5,height-themes.form.border.size,width-themes.form.border.size-5,height),Point(X,Y)) then cursor := crSizeS
  else if PtInRect(Rect(0,height-themes.form.border.size-5,themes.form.border.size+5,height),Point(X,Y)) then                              cursor := crSizeNE
  else cursor := crDefault;
end;

procedure TSForm.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
var
  i: integer;
begin
  inherited MouseDown(Button, Shift, X, Y);
  fdown   := true;
  GetCursorPos(fmouse);
  fbounds := Rect(left,top,left+width,top+height);
  fsizing := cursor <> crDefault;
  fmoving := not fsizing and PtInRect(Rect(themes.form.border.size,themes.form.border.size,width-themes.form.border.size,themes.form.border.size+themes.form.border.frame),Point(X,Y));
  for i:=0 to ControlCount-1 do Controls[i].Enabled:=false;
end;

procedure TSForm.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  p      : TPoint;
  l,t,w,h: integer;
begin
  inherited MouseMove(Shift, X, Y);

  if not fmoving and not fsizing then ThisCursor(X,Y);

  if fdown then begin

    GetCursorPos(p);

    if fmoving then begin
      left := fbounds.left +  p.x - fmouse.x;
      top  := fbounds.top  +  p.y - fmouse.y;
    end;

    if fsizing then begin

      l := left;
      t := top;
      w := width;
      h := height;

      case cursor of
        crSizeE  :
                   begin
                     l := fbounds.left +  p.x - fmouse.x;
                     w := (fbounds.right-fbounds.left) - (p.x - fmouse.x);
                   end;
        crSizeN  :
                   begin
                     t := fbounds.top  +  p.y - fmouse.y;
                     h := (fbounds.bottom-fbounds.top) - (p.y - fmouse.y);
                   end;
        crSizeW  : w := (fbounds.right-fbounds.left) + (p.x - fmouse.x);
        crSizeS  : h := (fbounds.bottom-fbounds.top) + (p.y - fmouse.y);

        crSizeSE :
                   begin
                     l := fbounds.left +  p.x - fmouse.x;
                     w := (fbounds.right-fbounds.left) - (p.x - fmouse.x);
                     t := fbounds.top  +  p.y - fmouse.y;
                     h := (fbounds.bottom-fbounds.top) - (p.y - fmouse.y);
                   end;

        crSizeSW :
                   begin
                     w := (fbounds.right-fbounds.left) + (p.x - fmouse.x);
                     t := fbounds.top  +  p.y - fmouse.y;
                     h := (fbounds.bottom-fbounds.top) - (p.y - fmouse.y);
                   end;

        crSizeNW :
                   begin
                     w := (fbounds.right-fbounds.left) + (p.x - fmouse.x);
                     h := (fbounds.bottom-fbounds.top) + (p.y - fmouse.y);
                   end;

        crSizeNE :
                   begin
                     h := (fbounds.bottom-fbounds.top) + (p.y - fmouse.y);
                     l := fbounds.left +  p.x - fmouse.x;
                     w := (fbounds.right-fbounds.left) - (p.x - fmouse.x);
                   end;
      end;

      if w < constraints.minwidth then begin
        w := constraints.minwidth;
        l := left;
      end;
      if h < constraints.minheight then begin
        h := constraints.minheight;
        t := top;
      end;

      SetBounds(l,t,w,h);
      {$IFDEF WINDOWS}Update;{$ENDIF}
    end;

  end;

end;

procedure TSForm.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i: integer;
begin
  inherited MouseUp(Button, Shift, X, Y);
  fdown   := false;
  fsizing := false;
  fmoving := false;
  ThisCursor(X,Y);
  for i:=0 to ControlCount-1 do Controls[i].Enabled:=true;
  Update;
end;
{$ENDIF}

procedure TSForm.SystemClick(Sender: TObject);
var
  i: integer;
begin
  case TSysMode(TSysButton(Sender).tag) of
    smClose  : Close;
    smReturn :
               begin
                 if fstatus = fsNormal then begin
                   fbounds := BoundsRect;
                   SetBounds(Screen.WorkAreaRect.left,Screen.WorkAreaRect.top,Screen.WorkAreaRect.right-Screen.WorkAreaRect.left,Screen.WorkAreaRect.bottom-Screen.WorkAreaRect.top);
                   fstatus := fsMaximized;
                 end else begin
                   SetBounds(fbounds.left,fbounds.top,fbounds.right-fbounds.left,fbounds.bottom-fbounds.top);
                   fstatus := fsNormal;
                   InvalidateRect(handle,nil,false);
                 end;
               end;
    smMinimal:
               begin
                 if fstatus = fsNormal then fbounds := BoundsRect;
                 for i := 0 to Application.ComponentCount - 1 do
                   if Application.Components[i] is TSForm then TSForm(Application.Components[i]).Hide;
                 ftray.Show;
               end;
    smToggle : fpanel.ShowFrame(ftoggle.Status);
  end;
end;

procedure TSForm.TrayClick(Sender: TObject);
begin
  Show;
  if fstatus = fsMaximized then begin
    SetBounds(Screen.WorkAreaRect.left,Screen.WorkAreaRect.top,Screen.WorkAreaRect.right-Screen.WorkAreaRect.left,Screen.WorkAreaRect.bottom-Screen.WorkAreaRect.top);
  end else begin
    SetBounds(fbounds.left,fbounds.top,fbounds.right-fbounds.left,fbounds.bottom-fbounds.top);
  end;
  ftray.Hide;
end;

procedure TSForm.Resize;
begin
  if handleallocated then begin
    if self <> Application.MainForm then begin
      if fpanel <> nil then FreeAndNil(fpanel);
      if ftoggle <> nil then FreeAndNil(ftoggle);
      if fminimal <> nil then FreeAndNil(fminimal);
    end else if fpanel <> nil then begin
      fpanel.SetBounds(themes.form.border.size,themes.form.border.size+themes.form.border.frame,width-themes.form.border.size*2,height-themes.form.border.size*2-themes.form.border.frame - (sbheight-themes.form.border.size));
    end;
    if floaded then begin
      fclose.left := width - fclose.width - themes.form.border.size;
      //fclose.color := themes.buttons.color;
      freturn.left := width - fclose.width - freturn.width - themes.form.border.size;
      //freturn.color := themes.buttons.color;
      if self = Application.MainForm then begin
        fminimal.left := width - fclose.width - freturn.width - fminimal.width - themes.form.border.size;
        //fminimal.color := themes.buttons.color;
      end;
    end else
      fbounds := BoundsRect;
    floaded := true;
  end;
end;

procedure TSForm.Paint;
begin
  canvas.StretchDraw(Rect(0,0,width,themes.form.border.size),fbitmap);
  canvas.StretchDraw(Rect(0,themes.form.border.size,themes.form.border.size,height-themes.form.border.size),fbitmap);
  canvas.StretchDraw(Rect(width-themes.form.border.size,themes.form.border.size,width,height-themes.form.border.size),fbitmap);
  canvas.StretchDraw(Rect(0,height-themes.form.border.size,width,height),fbitmap);

  canvas.StretchDraw(Rect(themes.form.border.size,themes.form.border.size,width-themes.form.border.size,themes.form.border.size+themes.form.border.frame),fbitmap);

  canvas.brush.style := bsClear;
  canvas.font.color  := themes.form.font.shadow;
  canvas.TextOut((width - tw) div 2 + 1,(themes.form.border.size + themes.form.border.frame - th - 1) div 2 + 1,caption);
  canvas.font.color  := themes.form.font.color;
  canvas.TextOut((width - tw) div 2,(themes.form.border.size + themes.form.border.frame - th - 1) div 2,caption);

  canvas.Draw(themes.form.border.size + 5,(themes.form.border.size + themes.form.border.frame - Application.Icon.Height) div 2, Application.Icon);
end;

begin
  screen := Forms.Screen;
end.

