unit SSynEdit;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, SThemes, Controls, Graphics, StdCtrls, Forms, SControls, LCLIntf,
  SynEdit, SynGutterLineNumber, SynGutter, SynGutterCodeFolding, SynGutterChanges,
  SynEditMarkupHighAll, SynEditHighlighter, SynEditMarkupWordGroup, SynEditMouseCmds,
  SynEditTypes, SynEditHighlighterFoldBase, SynOberon;

type

  { TSSynEdit }

  TSSynEdit = class(TCustomControl)
  private
    feditor     : TSynEdit;
    fvscroll    : TSScrollBar;
    fhscroll    : TSScrollBar;
    fword       : string;
    function    GetEditor: TSynEdit;
    function    GetHighlighter: TSynCustomHighlighter;
    procedure   SetEditor(AValue: TSynEdit);
    procedure   SetHighlighter(AValue: TSynCustomHighlighter);
    procedure   SwapCommand(idx1,idx2: TSynEditorMouseCommand);
    procedure   EditorKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure   EditorKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure   EditorMouseLink(Sender: TObject; X, Y: Integer; var AllowMouseLink: Boolean);
    procedure   EditorClickLink(Sender: TObject; Button: TMouseButton;Shift: TShiftState; X, Y: Integer);
    procedure   ScrollChange(Sender: TObject);
    procedure   EditPaint(Sender: TObject; ACanvas: TCanvas);
    procedure   EditMouseEvent(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

  public
    encode    : string;
    ErrorLine : integer;
    DebugLine : integer;
    SynMarkup : TSynEditMarkupHighlightAllCaret;
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   LoadFromFile(const FileName: string);
    property    Editor : TSynEdit read GetEditor write SetEditor;
    property    Highlighter: TSynCustomHighlighter read GetHighlighter write SetHighlighter;
  published

  end;

implementation

var
  oberonhlg : TSynCustomFoldHighlighter;

{ TSSynEdit }

constructor TSSynEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  feditor  := TSynEdit.Create(self);
  feditor.Parent := self;
  with feditor do begin
    align         := alClient;
    Options       := Options + [eoDragDropEditing];
    //Font.Name   := 'PT Mono';
    //Font.Size   := 10;

    Highlighter := oberonhlg;
    OnMouseDown := @EditMouseEvent;
    OnMouseUp   := @EditMouseEvent;

    Font.Color  := themes.editor.font.color;
    Font.Quality:= fqNonAntialiased;
    BorderStyle := bsNone;
    ScrollBars  := ssNone;

    LineHighlightColor.Background := themes.editor.line;
    SelectedColor.Background      := themes.editor.select;
    color                         := themes.editor.color;
    RightEdgeColor                := themes.editor.right;
    Gutter.Color                  := themes.editor.gutter;
    BracketMatchColor.Foreground  := themes.editor.matches;
    BracketMatchColor.style       := [];

    TSynGutterLineNumber(Gutter.Parts.Part[1]).DigitCount:=3;
    TSynGutterLineNumber(Gutter.Parts.Part[1]).ShowOnlyLineNumbersMultiplesOf:=10;
    TSynGutterLineNumber(Gutter.Parts.Part[1]).MarkupInfo.Foreground:=themes.editor.number; //Цыфры
    TSynGutterLineNumber(Gutter.Parts.Part[1]).MarkupInfo.Background:=themes.editor.gutter;
    TSynGutterChanges(Gutter.Parts.Part[2]).ModifiedColor := themes.editor.changes.modifed;
    TSynGutterChanges(Gutter.Parts.Part[2]).SavedColor := themes.editor.changes.saved;
    TSynGutterSeparator(Gutter.Parts.Part[3]).MarkupInfo.Foreground :=themes.editor.gutter;  //Разделитель
    TSynGutterSeparator(Gutter.Parts.Part[3]).MarkupInfo.Background :=themes.editor.gutter;
    TSynGutterCodeFolding(Gutter.Parts.Part[4]).MarkupInfo.Foreground:=themes.editor.folding;
    TSynGutterCodeFolding(Gutter.Parts.Part[4]).MarkupInfo.Background:=themes.editor.gutter;
    TSynEditMarkupWordGroup(MarkupByClass[TSynEditMarkupWordGroup]).Enabled := false;
    FoldedCodeColor.FrameColor:=0;

    OnMouseLink := @EditorMouseLink;
    OnClickLink := @EditorClickLink;
    OnKeyDown   := @EditorKeyDown;
    OnKeyUp     := @EditorKeyUp;
    OnPaint     := @EditPaint;

    MouseLinkColor.Background := clNone;
    MouseLinkColor.Foreground := clNone;
    MouseLinkColor.FrameColor := $FF5555;
    MouseLinkColor.FrameEdges := sfeBottom;
    MouseLinkColor.FrameStyle := slsDotted;
    MouseLinkColor.Style:=[];

    MouseActions.Add;
    MouseActions.Items[0].Shift        := [ssCtrl];
    MouseActions.Items[0].ShiftMask    := [ssShift, ssAlt, ssCtrl];

    Text := ' ';
  end;

  SynMarkup := TSynEditMarkupHighlightAllCaret(feditor.MarkupByClass[TSynEditMarkupHighlightAllCaret]);
  SynMarkup.MarkupInfo.FrameColor := themes.editor.markup;
  SynMarkup.MarkupInfo.Background := themes.editor.markup;
  SynMarkup.WaitTime := 20;
  SynMarkup.Trim := True;
  SynMarkup.FullWord := True;
  SynMarkup.IgnoreKeywords := True;

  fvscroll        := TSScrollBar.Create(self);
  fvscroll.Parent := self;
  fvscroll.align  := alRight;
  fvscroll.kind   := sbVertical;
  fvscroll.OnChange := @ScrollChange;

  fhscroll        := TSScrollBar.Create(self);
  fhscroll.Parent := self;
  fhscroll.align  := alBottom;
  fhscroll.OnChange := @ScrollChange;
  fhscroll.shift  := fvscroll.width;
end;

destructor TSSynEdit.Destroy;
begin
  FreeAndNil(fvscroll);
  FreeAndNil(feditor);
  inherited Destroy;
end;

function TSSynEdit.GetEditor: TSynEdit;
begin
  Result := feditor;
end;

procedure TSSynEdit.SetEditor(AValue: TSynEdit);
begin
  if feditor <> AValue then feditor := AValue;
end;

function TSSynEdit.GetHighlighter: TSynCustomHighlighter;
begin
  Result := feditor.Highlighter;
end;

procedure TSSynEdit.SetHighlighter(AValue: TSynCustomHighlighter);
begin
  if feditor.Highlighter <> AValue then feditor.Highlighter := AValue;
end;

procedure TSSynEdit.SwapCommand(idx1,idx2: TSynEditorMouseCommand);
var
  i: integer;
begin
  for i:= 0 to feditor.MouseActions.Count - 1  do
    if feditor.MouseActions.Items[i].Command = idx1 then
      feditor.MouseActions.Items[i].Command := idx2;
end;

procedure TSSynEdit.EditorKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if ssCtrl IN Shift then begin
    SwapCommand(emcNone,emcMouseLink);
    feditor.MouseOptions := [emUseMouseActions,emShowCtrlMouseLinks];
  end;
  InvalidateRect(handle,nil,false);
end;

procedure TSSynEdit.EditorKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if key=17 then begin
    SwapCommand(emcMouseLink,emcNone);
    feditor.MouseOptions := [];
  end;
end;

procedure TSSynEdit.EditorClickLink(Sender: TObject; Button: TMouseButton;Shift: TShiftState; X, Y: Integer);
begin
  SwapCommand(emcMouseLink,emcNone);
  feditor.MouseOptions := [];
end;

procedure TSSynEdit.EditorMouseLink(Sender: TObject; X, Y: Integer;var AllowMouseLink: Boolean);
var
  Attri: TSynHighlighterAttributes;
  toktype,start: integer;
begin
  AllowMouseLink := feditor.GetHighlighterAttriAtRowColEx(Point(X,Y),fword,toktype,start,Attri) and (toktype = 3);
end;

procedure TSSynEdit.ScrollChange(Sender: TObject);
begin
  feditor.TopLine  := fvscroll.Position+1;
  feditor.LeftChar := fhscroll.Position;
end;

procedure TSSynEdit.EditPaint(Sender: TObject; ACanvas: TCanvas);
begin
  fvscroll.Max      := feditor.Lines.Count - 1;
  fvscroll.PageSize := feditor.Height div feditor.LineHeight;
  fvscroll.Position := feditor.TopLine - 1;

  fhscroll.Max      := feditor.MaxLeftChar;
  fhscroll.PageSize := feditor.CharsInWindow;
  fhscroll.Position := feditor.LeftChar;
end;

procedure TSSynEdit.EditMouseEvent(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if assigned(feditor.OnChange) then feditor.OnChange(Sender);
end;

procedure TSSynEdit.LoadFromFile(const FileName: string);
begin
  feditor.Lines.LoadFromFile(FileName);
  if (feditor.Lines.Text = '') then feditor.Lines.Text := ' ';
end;

initialization
  oberonhlg := TSynOberon.Create(nil);

finalization
  FreeAndNil(oberonhlg);
end.

