unit STabs;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Controls, LCLIntf, Graphics, SThemes,
  ExtCtrls, Dialogs;

type

  TSTabControl = class;

  { TSTab }

  TSTab = class
    parent    : TSTabControl;
    bounds    : TRect;
    realrect  : TRect;
    textpos   : TPoint;
    realtext  : TPoint;
    width     : integer;
    undomouse : boolean;
    enterclose: boolean;
    leaveclose: boolean;
    undoclose : boolean;
    active    : boolean;
    caption   : string;
    path      : string;
    control   : TCustomControl;
    index     : integer;
    loaded    : boolean;
    fmodified : boolean;
  private
    procedure SetModified(AValue: boolean);
  public
    property    modified: boolean read fmodified write SetModified;
  end;

  TSPageControl = class;

  { TSTabControl }
  TCloseEvent = procedure(Sender: TObject; var CanClose: boolean) of object;

  TSTabControl = class(TCustomControl)
  protected
    procedure MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseLeave; override;
    procedure Resize; override;
    procedure Paint; override;
  private
    { private declarations }
    factive   : TSTab;
    ftabs     : TStringList;
    fdown     : boolean;
    fclose    : boolean;
    fclosenum : integer;
    fmouse    : TRect;
    fcontrol  : TCustomControl;
    fstart    : integer;
    fend      : integer;
    foffset   : integer;
    ftableft  : boolean;
    ftabright : boolean;
    ftableave : boolean;
    fonchange : TNotifyEvent;
    fonclose  : TCloseEvent;
    function    AddTab(const txt: string):TSTab;
    function    AddNew(const txt: string):TSTab;
    procedure   DelTab(n,x,y: integer);
    procedure   CalcEnd;
  public
    { public declarations }
    constructor Create(TheOwner: TSPageControl); overload;
    destructor  Destroy; override;
    property    Control : TCustomControl read fcontrol;
    property    OnChange: TNotifyEvent read fonchange write fonchange;
    property    OnClose : TCloseEvent read fonclose write fonclose;
  end;

  { TSPageControl }

  TSPageControl = class(TCustomControl)
  protected
  private
    { private declarations }
    ftabs       : TSTabControl;
    fonchange   : TNotifyEvent;
    fonclose    : TCloseEvent;
    function    GetCount: integer;
  public
    { public declarations }
    constructor Create(TheOwner: TComponent);override;
    destructor  Destroy; override;
    function    AddTab(const txt: string):TSTab;
    function    AddNew(const txt: string):TSTab;
    function    GetBy(const n: integer):TSTab;
    procedure   Change(Sender:TObject);
    procedure   TabClose(Sender:TObject; var done: boolean);
    property    Count : integer read GetCount;
    property    OnChange: TNotifyEvent read fOnChange write fOnChange;
    property    OnClose : TCloseEvent read fonclose write fonclose;
  end;

implementation

{ TSTab }

procedure TSTab.SetModified(AValue: boolean);
begin
  if fmodified=AValue then Exit;
  fmodified:=AValue;
  InvalidateRect(parent.handle,nil,false);
end;

{ TSTabControl }

constructor TSTabControl.Create(TheOwner: TSPageControl);
begin
  inherited Create(TheOwner);
  width := 250;
  height := 30;
  color := themes.form.color;
  ftabs := TStringList.Create;
  fcontrol := nil;
  fstart   := 0;
  fend     := 0;
  foffset  := 0;
  ftableft := false;
  ftabright:= false;
  ftableave:= false;
  fclosenum:= -1;
end;

destructor TSTabControl.Destroy;
var
  i: integer;
begin
  for i:=0 to ftabs.count - 1 do ftabs.objects[i].free;
  FreeAndNil(ftabs);
  inherited Destroy;
end;

procedure TSTabControl.MouseLeave;
var
  i   : integer;
begin
  inherited MouseLeave;
  for i:=0 to ftabs.count - 1 do
    TSTab(ftabs.objects[i]).undomouse := false;
  InvalidateRect(handle,nil,false);
end;

procedure TSTabControl.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
var
  i   : integer;
begin
  inherited MouseDown(Button, Shift, X, Y);

  fdown := Button = mbLeft;
  ftableft  := PtInRect(Rect(width - 50, 4, width - 27,height - 4),Point(X,Y));
  ftabright := PtInRect(Rect(width - 27, 4, width - 5,height - 4),Point(X,Y));
  if ftableft or ftabright then begin
    InvalidateRect(handle,nil,false);
    exit;
  end;

  if fdown then begin

    fclosenum := -1;
    for i:=0 to ftabs.count - 1 do begin
      fclose := PtInRect(Rect(TSTab(ftabs.objects[i]).bounds.right-foffset-20,TSTab(ftabs.objects[i]).bounds.top+10,TSTab(ftabs.objects[i]).bounds.right-foffset-10,TSTab(ftabs.objects[i]).bounds.bottom-10),Point(X,Y)) and TSTab(ftabs.objects[i]).undomouse;
      if fclose then begin
        fclosenum := i;
        exit;
      end;
    end;

    factive := nil;
    if (ftabs.count>0) and (X < TSTab(ftabs.objects[fend - 1]).bounds.right - foffset) then begin
      for i:=0 to ftabs.count - 1 do begin
        TSTab(ftabs.objects[i]).active := TSTab(ftabs.objects[i]).undomouse;
        TSTab(ftabs.objects[i]).control.visible := TSTab(ftabs.objects[i]).active;
        if TSTab(ftabs.objects[i]).undomouse then begin
          factive := TSTab(ftabs.objects[i]);
          factive.realrect := factive.bounds;
          factive.realtext := factive.textpos;
          fmouse.left := X - factive.bounds.left;
          fmouse.right := X - factive.bounds.right;
          fmouse.top   := X - factive.textpos.x;
          fmouse.bottom:= X - factive.bounds.left + foffset;
        end;
      end;
      InvalidateRect(handle,nil,false);
    end;

  end;

  if Assigned(fOnChange) then fOnChange(factive);

end;

procedure TSTabControl.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  tab,t : TSTab;
  i,w   : integer;
  found : boolean;
  found2: boolean;

  procedure SwapTab(i1,i2: integer; moveleft: boolean);
  var
    t1,t2 : TSTab;
    r1    : integer;
  begin
    if (i1>=0) and (i2>=0) and (i1<=ftabs.count-1) and (i2<=ftabs.count-1) then begin
      t1 := TSTab(ftabs.Objects[i1]);
      t2 := TSTab(ftabs.Objects[i2]);

      if moveleft then begin
        r1 := t2.bounds.left;
        t2.bounds.left  := t2.bounds.left + t1.width;
        t2.bounds.right := t2.bounds.left + t2.width;
      end else begin
        r1 := t1.realrect.left + t2.width;
        t2.bounds.left  := t1.realrect.left;
        t2.bounds.right := t2.bounds.left + t2.width;
      end;

      t2.textpos.x      := t2.bounds.left + 10;
      t2.realtext       := t2.textpos;
      t2.realrect       := t2.bounds;

      t1.realrect.left  := r1;
      t1.realrect.right := t1.realrect.left + t1.width;
      t1.realtext.x     := t1.realrect.left + 10;

      t1.index := i2;
      t2.index := i1;
      ftabs.Exchange(i1,i2);
    end;
  end;

begin
  inherited MouseMove(Shift, X, Y);

  if fclose then exit;

  if PtInRect(Rect(width - 50, 4, width - 5,height - 4),Point(X,Y)) then begin
    ftableft  := PtInRect(Rect(width - 50, 4, width - 27,height - 4),Point(X,Y));
    ftabright := PtInRect(Rect(width - 27, 4, width - 5,height - 4),Point(X,Y));
    InvalidateRect(handle,nil,false);
    ftableave := true;
    exit;
  end else if ftableave and not fdown then begin
    ftableft  := false;
    ftabright := false;
    InvalidateRect(handle,nil,false);
    ftableave := false;
  end;

  if fdown and (ftabs.count > 1) then begin
    if factive <> nil then begin

      factive.bounds.left  := X - fmouse.bottom + foffset;
      factive.bounds.right := factive.bounds.left + factive.width;
      factive.textpos.x    := factive.bounds.left + 10;

      while (X - fmouse.bottom < 0) and (fstart>=0) do begin
        if fstart > 0 then begin
          dec(fstart);
          CalcEnd;
          SwapTab(factive.index,factive.index-1,true);
          factive.bounds.left  := X - fmouse.bottom + foffset;
          factive.bounds.right := factive.bounds.left + factive.width;
          factive.textpos.x    := factive.bounds.left + 10;
        end else if factive.bounds.left < 0 then begin
          factive.bounds.left  := 0;
          factive.bounds.right := factive.bounds.left + factive.width;
          factive.textpos.x    := factive.bounds.left + 10;
        end;
        InvalidateRect(handle,nil,false);
        exit;
      end;

      w := 0;
      for i:=fstart to fend-1 do w := w + TSTab(ftabs.Objects[i]).width;

      while (factive.index <= ftabs.count - 1) and (factive.bounds.right - foffset > w) do begin
        if fend < ftabs.count then begin
          inc(fstart);
          CalcEnd;
          SwapTab(factive.index,factive.index+1,false);
          factive.bounds.left  := X - fmouse.bottom + foffset;
          factive.bounds.right := factive.bounds.left + factive.width;
          factive.textpos.x    := factive.bounds.left + 10;
        end else begin
          factive.bounds.left  := TSTab(ftabs.Objects[factive.index-1]).realrect.right;
          factive.bounds.right := factive.bounds.left + factive.width;
          factive.textpos.x    := factive.bounds.left + 10;
        end;
        InvalidateRect(handle,nil,false);
        exit;
      end;

      i := (factive.bounds.right - factive.bounds.left) div 2;
      if (factive.index = fend - 1) and (factive.bounds.left > TSTab(ftabs.Objects[factive.index-1]).bounds.right) then begin
        factive.bounds  := factive.realrect;
        factive.textpos := factive.realtext;
      end else if (factive.index > 0) and (factive.bounds.left < TSTab(ftabs.Objects[factive.index-1]).bounds.right - TSTab(ftabs.Objects[factive.index-1]).width div 2) then begin
        SwapTab(factive.index,factive.index-1,true);
      end else if (factive.index < ftabs.count - 1) and (factive.bounds.right > TSTab(ftabs.Objects[factive.index+1]).bounds.left + TSTab(ftabs.Objects[factive.index+1]).width div 2) then
        SwapTab(factive.index,factive.index+1,false);

      InvalidateRect(handle,nil,false);
    end;
    exit;
  end;

  found := false;
  found2 := false;
  tab := nil;
  for i:=0 to ftabs.count - 1 do begin
    tab := TSTab(ftabs.objects[i]);
    if tab.undomouse then t := tab;
    tab.undomouse := PtInRect(Rect(tab.bounds.left-foffset,tab.bounds.top,tab.bounds.right-foffset,tab.bounds.bottom),Point(X,Y));
    if tab.undomouse and (t<>tab) then found := true;

    tab.leaveclose := tab.undomouse;
    if tab.undomouse then begin
       tab.enterclose := PtInRect(Rect(tab.bounds.right-foffset-20,tab.bounds.top+10,tab.bounds.right-foffset-10,tab.bounds.bottom-10),Point(X,Y));
       tab.undoclose  := tab.enterclose and tab.leaveclose;
       if tab.undoclose then found := true;
       found2 := not tab.enterclose and tab.leaveclose
    end;

  end;
  if not found and (tab<>nil) and (tab.bounds.right-foffset<X) then found := true;
  if found or found2 then InvalidateRect(handle,nil,false);
end;

procedure TSTabControl.MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
var
  i   : integer;
  r   : TRect;
  done: boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);

  ftableft  := PtInRect(Rect(width - 50, 4, width - 27,height - 4),Point(X,Y));
  ftabright := PtInRect(Rect(width - 27, 4, width - 5,height - 4),Point(X,Y));

  if ftableft  and (fend < ftabs.count) and (factive = nil) then inc(fstart);
  if ftabright and (fstart > 0) and (factive = nil) then dec(fstart);
  if ftableft or ftabright and (factive = nil) then begin
    InvalidateRect(handle,nil,false);
    fdown := false;
    CalcEnd;
    for i:=0 to ftabs.count - 1 do TSTab(ftabs.objects[i]).undomouse := false;
    InvalidateRect(handle,nil,false);
  end;

  if fdown then begin
    fdown := false;

    if fclose and (fclosenum >- 1) then begin
      done := true;
      if assigned(fonclose) then fonclose(TSTab(ftabs.objects[fclosenum]),done);
      if done then DelTab(fclosenum,X,Y);
    end;
    fclose := false;

    if factive <> nil then begin
      factive.bounds  := factive.realrect;
      factive.textpos := factive.realtext;
      InvalidateRect(handle,nil,false);
      factive := nil;
    end;

    if (ftabs.count>0) and (X < TSTab(ftabs.objects[ftabs.count - 1]).bounds.right - foffset) then begin
      for i:=0 to ftabs.count - 1 do begin
        r := TSTab(ftabs.objects[i]).bounds;
        dec(r.left,foffset);
        dec(r.right,foffset);
        TSTab(ftabs.objects[i]).undomouse := PtInRect(r,Point(X,Y));
        TSTab(ftabs.objects[i]).undoclose := TSTab(ftabs.objects[i]).undomouse and PtInRect(Rect(TSTab(ftabs.objects[i]).bounds.right-foffset-20,TSTab(ftabs.objects[i]).bounds.top+10,TSTab(ftabs.objects[i]).bounds.right-foffset-10,TSTab(ftabs.objects[i]).bounds.bottom-10),Point(X,Y)) and TSTab(ftabs.objects[i]).undomouse;
      end;
      InvalidateRect(handle,nil,false);
    end;
  end;

end;

function TSTabControl.AddTab(const txt: string):TSTab;
var
  th  : integer;
  i,j : integer;
  done: boolean;
  path,capt: string;
begin

  path  := ExtractFilePath(txt);
  capt  := ExtractFileName(txt);
  i := 0;
  while (i<ftabs.count) do begin
    if (TSTab(ftabs.objects[i]).caption = capt) and (TSTab(ftabs.objects[i]).path = path) then begin
      for j:=0 to ftabs.count - 1 do begin
        TSTab(ftabs.objects[j]).active := TSTab(ftabs.objects[j]) = TSTab(ftabs.objects[i]);
        TSTab(ftabs.objects[j]).control.visible := TSTab(ftabs.objects[j]).active;
        if TSTab(ftabs.objects[j]).active then fcontrol := TSTab(ftabs.objects[j]).control;
      end;
      TSTab(ftabs.objects[i]).loaded := true;
      result := TSTab(ftabs.objects[i]);
      if Assigned(fOnChange) then fOnChange(result);
      exit;
    end;
    inc(i);
  end;

  th := canvas.textheight('X');
  result := TSTab.Create;
  result.parent  := self;
  result.path    := ExtractFilePath(txt);
  result.caption := ExtractFileName(txt);
  result.undomouse := false;

  result.control   := TCustomControl.Create(self);
  result.control.parent := Parent;
  result.control.borderwidth := 5;
  result.control.align := alClient;
  result.control.childsizing.topbottomspacing:=1;
  result.control.childsizing.leftrightspacing:=1;

  result.bounds.top  := 0;
  if ftabs.count = 0 then
    result.bounds.left := 0
  else
    result.bounds.left := TSTab(ftabs.objects[ftabs.count-1]).bounds.right;
  result.bounds.right := result.bounds.left + 10 + canvas.textwidth(result.caption) + 10 + 16;
  result.bounds.bottom := height;
  result.textpos.x := result.bounds.left + 10;
  result.textpos.y := (result.bounds.bottom - result.bounds.top - th) div 2;
  result.index := ftabs.count;
  result.realrect := result.bounds;
  result.realtext := result.textpos;
  result.width    := result.bounds.right - result.bounds.left;
  result.loaded   := false;
  result.fmodified := false;
  ftabs.AddObject(txt,result);
  for i:=0 to ftabs.count - 1 do begin
    TSTab(ftabs.objects[i]).active := TSTab(ftabs.objects[i]) = result;
    TSTab(ftabs.objects[i]).control.visible := TSTab(ftabs.objects[i]).active;
    if TSTab(ftabs.objects[i]).active then begin
      fcontrol := TSTab(ftabs.objects[i]).control;
    end;
  end;
  if Assigned(fOnChange) then fOnChange(result);
  color := themes.form.border.color;
  done := false;
  while result.bounds.right - foffset > width - 55 do begin
    inc(fstart);
    CalcEnd;
    if not done then done := true;
  end;
  if not done then CalcEnd;
end;

function TSTabControl.AddNew(const txt: string):TSTab;
var
  i,j: integer;
  n,path,capt,short,ext: string;
  done: boolean;
begin
  n := txt;
  path  := ExtractFilePath(txt);
  capt  := ExtractFileName(txt);
  ext   := ExtractFileExt(capt);
  short := copy(capt,1,pos(ext,capt)-1);

  if (ftabs.count>0) then begin
    done := true;
    j := 1;
    while done do begin
      i:=0;
      done := false;
      while (i<ftabs.count) and not done do begin
        n := short+IntToStr(j)+ext;
        done := (TSTab(ftabs.objects[i]).caption = n);
        inc(i);
      end;
      inc(j);
    end;
  end else
    n := short+'1'+ext;
  result := AddTab(n);
end;

procedure TSTabControl.DelTab(n,x,y: integer);
var
  i    : integer;
  r    : TRect;
  found: boolean;
begin

  if PtInRect(Rect(TSTab(ftabs.objects[n]).bounds.right-foffset-20,TSTab(ftabs.objects[n]).bounds.top+10,TSTab(ftabs.objects[n]).bounds.right-foffset-10,TSTab(ftabs.objects[n]).bounds.bottom-10),Point(X,Y)) then begin

    found := false;

    for i:=0 to ftabs.count - 1 do begin
      if i>=n+1 then begin
        r := TSTab(ftabs.objects[i]).bounds;
        dec(r.left,TSTab(ftabs.objects[n]).bounds.right-TSTab(ftabs.objects[n]).bounds.left);
        dec(r.right,TSTab(ftabs.objects[n]).bounds.right-TSTab(ftabs.objects[n]).bounds.left);
        TSTab(ftabs.objects[i]).bounds := r;
        TSTab(ftabs.objects[i]).textpos.x := TSTab(ftabs.objects[i]).bounds.left + 10;
        TSTab(ftabs.objects[i]).index := i;
        TSTab(ftabs.objects[i]).realrect := TSTab(ftabs.objects[i]).bounds;
        TSTab(ftabs.objects[i]).realtext := TSTab(ftabs.objects[i]).textpos;
      end;
      if TSTab(ftabs.objects[i]).active and not TSTab(ftabs.objects[n]).active then found := true;
    end;

    TSTab(ftabs.objects[n]).control.Free;
    TSTab(ftabs.objects[n]).Free;
    ftabs.Delete(n);
    CalcEnd;

    if not found and (ftabs.count > 0) then begin
      if n < ftabs.count then begin
        TSTab(ftabs.objects[n]).active := true;
        TSTab(ftabs.objects[n]).control.visible := true;
        if Assigned(fOnChange) then fOnChange(TSTab(ftabs.objects[n]))
      end else if n > 0 then begin
        TSTab(ftabs.objects[n-1]).active := true;
        TSTab(ftabs.objects[n-1]).control.visible := true;
        if Assigned(fOnChange) then fOnChange(TSTab(ftabs.objects[n-1]));
      end;
    end;

    if Assigned(fOnChange) and (ftabs.count = 0) then fOnChange(nil);

    fclosenum := -1;
    InvalidateRect(handle,nil,false);

  end else begin
    TSTab(ftabs.objects[n]).undoclose := false;
    TSTab(ftabs.objects[n]).undomouse := false;
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSTabControl.Resize;
begin
  if handleallocated then begin
    if (fstart > 0) and (ftabs.count > 0) and (TSTab(ftabs.objects[fstart-1]).width + TSTab(ftabs.objects[ftabs.count-1]).bounds.right - foffset < width - 55) then dec(fstart);
    CalcEnd;
    InvalidateRect(handle,nil,false);
  end;
end;

procedure TSTabControl.CalcEnd;
var
  i,w: integer;
begin
  fend := fstart;
  if ftabs.count > fstart then begin
    w := 0;
    while (fend < ftabs.count) and (w + TSTab(ftabs.objects[fend]).width < width - 55) do begin
      w := w + TSTab(ftabs.objects[fend]).width;
      inc(fend);
    end;
    if fend > ftabs.count then fend := ftabs.count;
  end;
  foffset := 0;
  if fstart < ftabs.count then
    for i := 0 to fstart - 1 do foffset := foffset + TSTab(ftabs.objects[i]).width;
end;

procedure TSTabControl.Paint;
var
  i,closeclr: integer;
  tab,t : TSTab;

  procedure DrawCloseBtn(cx,cy,clr: integer);
  begin
    canvas.pen.width := 2;
    canvas.pen.color := clr;
    canvas.Line(cx-4,cy-4,cx+4,cy+4);
    canvas.Line(cx+4,cy-4,cx-4,cy+4);
  end;

  procedure DrawModified(r: TRect; clr: integer);
  begin
    canvas.pen.width   := 1;
    canvas.pen.color   := clr;
    canvas.brush.color := clr;
    canvas.line(r.left+30,r.bottom-4,r.right-30,r.bottom-4);
    canvas.line(r.left+30,r.bottom-4-1,r.right-30,r.bottom-4-1);
  end;

begin
  t := nil;

  if (fstart = fend) and (ftabs.count > 0) then fend := fstart + 1;

  for i := fstart to fend - 1 do begin
    tab := TSTab(ftabs.objects[i]);
    canvas.font.color := themes.tabs.font.shadow;
    canvas.brush.style := bsSolid;
    if tab.undomouse then begin
      canvas.pen.color := themes.buttons.light;
      canvas.brush.color := themes.buttons.light;
      canvas.font.color := themes.tabs.font.light;
      canvas.rectangle(tab.bounds.left - foffset,tab.bounds.top,tab.bounds.right - foffset,tab.bounds.bottom);
    end;
    if tab.active then t := tab;
    canvas.brush.style := bsClear;
    canvas.textout(tab.textpos.x - foffset,tab.textpos.y,tab.caption);
    if tab.undoclose then
      closeclr := $b8b8ff
    else
      closeclr := canvas.font.color;
    DrawCloseBtn(tab.bounds.right - foffset - 14,(tab.bounds.bottom - tab.bounds.top) div 2,closeclr);
    if tab.fmodified then DrawModified(tab.bounds,canvas.font.color);
  end;
  if t<>nil then begin
    canvas.pen.color := themes.tabs.color;
    canvas.brush.color := themes.tabs.color;
    canvas.font.color := themes.tabs.font.color;
    canvas.rectangle(t.bounds.left - foffset,t.bounds.top,t.bounds.right - foffset,t.bounds.bottom);
    canvas.brush.style := bsClear;
    canvas.textout(t.textpos.x - foffset,t.textpos.y,t.caption);
    if t.undoclose then
      closeclr := $8080ff
    else
      closeclr := canvas.font.color;
    DrawCloseBtn(t.bounds.right - foffset - 14,(t.bounds.bottom - tab.bounds.top) div 2,closeclr);
    if t.fmodified then DrawModified(t.bounds,themes.tabs.modified);
  end;
  if (fstart > 0) or (fend < ftabs.count) then begin
    canvas.pen.width   := 1;
    canvas.pen.color := themes.buttons.hight;
    canvas.brush.color := themes.buttons.light;
    canvas.brush.style := bsSolid;
    canvas.roundrect(width - 50,4,width - 4 ,height-4,20,20);
    canvas.line(width - 27,4,width - 27,height - 4);

    canvas.pen.color := themes.buttons.every;
    canvas.brush.color := themes.buttons.every;
    if fend < ftabs.count then begin
      canvas.pen.color := clWhite;
      canvas.brush.color := clWhite;
      if ftableft then begin
        canvas.pen.color := themes.tabs.arrow;
        canvas.brush.color := themes.tabs.arrow;
      end;
    end;
    canvas.polygon([Point(width - 41,height div 2),Point(width - 37,11),Point(width - 37,19)]);

    canvas.pen.color := themes.buttons.every;
    canvas.brush.color := themes.buttons.every;
    if fstart > 0 then begin
      canvas.pen.color := clWhite;
      canvas.brush.color := clWhite;
      if ftabright then begin
        canvas.pen.color := themes.tabs.arrow;
        canvas.brush.color := themes.tabs.arrow;
      end;
    end;
    canvas.polygon([Point(width - 13,height div 2),Point(width - 17,11),Point(width - 17,19)]);
  end;
end;

{ TSPageControl }

function TSPageControl.GetCount: integer;
begin
  result := ftabs.ftabs.count;
end;

constructor TSPageControl.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  width := 200;
  height := 200;
  ftabs := TSTabControl.Create(self);
  ftabs.parent   := self;
  ftabs.align    := alTop;
  ftabs.OnChange := @Change;
  ftabs.OnClose  := @TabClose;
end;

destructor TSPageControl.Destroy;
begin
  FreeAndNil(ftabs);
  inherited Destroy;
end;

function TSPageControl.AddTab(const txt: string):TSTab;
begin
  result := ftabs.AddTab(txt);
  InvalidateRect(ftabs.handle,nil,false);
end;

function TSPageControl.AddNew(const txt: string):TSTab;
begin
  result := ftabs.AddNew(txt);
  InvalidateRect(ftabs.handle,nil,false);
end;

function TSPageControl.GetBy(const n: integer):TSTab;
begin
  result := TSTab(ftabs.ftabs.objects[n]);
end;

procedure TSPageControl.Change(Sender:TObject);
begin
  if Assigned(fOnChange) then fOnChange(TSTab(Sender));
end;

procedure TSPageControl.TabClose(Sender:TObject; var done: boolean);
begin
  if Assigned(fonclose) then fonclose(TSTab(Sender),done);
end;

end.

