unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, SForms, STabs, Unit2, StdCtrls, ExtCtrls, Controls, Graphics, Dialogs,
  SThemes, SControls, SDialogs,
  SSynEdit, SynEdit, Forms;

type
  { TForm1 }

  TForm1 = class(TSForm)
    Edit32      : TImageList;
    File16      : TImageList;
    File32      : TImageList;
    Edit16      : TImageList;
    OpenDialog1 : TOpenDialog;
    Other32     : TImageList;
    View32      : TImageList;
    procedure FormCloseQuery(Sender: TObject; var CanClose: boolean);
    procedure   FormCreate(Sender: TObject);
  private
    { private declarations }
    statusbar   : TStatusBar;
    factive     : TSTab;
    feditor     : TSynEdit;
    fopendialog : TSOpenDialog;
    fsavedialog : TSSaveDialog;
    procedure   SetEdit(var se: TSSynEdit; tb: TSTab);
    procedure   NewFile();
    procedure   OpenFile();
    function    SaveFileAs(): boolean;
    function    SaveFile(): boolean;
    function    SaveAll(): boolean;
    function    CloseFile(): boolean;
    function    CloseAll(): boolean;
    procedure   PrintFile();
  public
    { public declarations }
    tab         : TSPageControl;
    se          : TSSynEdit;
    destructor  Destroy; override;
    procedure   TabChange(Sender:TObject);
    procedure   EditorChange(Sender:TObject);
    procedure   TabClose(Sender:TObject; var done: boolean);
    procedure   MenuClick(Sender: TObject; btnindex: integer; menuindex: integer);
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
var
  btn : TPanelButton;
begin
  btn := SysPanel.AddButton('File');
  btn.ImageBig    := File32;
  btn.ImageSmall  := File16;
  btn.DefaultIcon := 0;
  btn.OnClick:=@MenuClick;
  btn.AddMenu('New',0);
  btn.AddMenu('Open',1);
  btn.AddMenu('-');
  btn.AddMenu('Save',2);
  btn.AddMenu('Save as...',3);
  btn.AddMenu('Save all...',4);
  btn.AddMenu('-');
  btn.AddMenu('Print',5);
  btn.AddMenu('-');
  btn.AddMenu('Exit',6);
  btn.OnMenuClick := @MenuClick;

  btn := SysPanel.AddButton('Edit');
  btn.ImageBig    := Edit32;
  btn.ImageSmall  := Edit16;
  btn.DefaultIcon := 9;
  btn.OnClick     := @MenuClick;
  btn.AddMenu('Undo',0);
  btn.AddMenu('Redo',1);
  btn.AddMenu('-');
  btn.AddMenu('Cut',2);
  btn.AddMenu('Copy',3);
  btn.AddMenu('Paste',4);
  btn.AddMenu('-');
  btn.AddMenu('Find',5);
  btn.AddMenu('Find previos',6);
  btn.AddMenu('Find next',7);
  btn.AddMenu('Replace',8);
  btn.OnMenuClick := @MenuClick;

  btn := SysPanel.AddButton('View');
  btn.ImageBig    := View32;
  btn.DefaultIcon := 0;
  btn.OnClick:=@MenuClick;

  btn := SysPanel.AddButton('Help');
  btn.ImageBig    := Other32;
  btn.DefaultIcon := 0;
  btn.OnClick:=@MenuClick;

  btn := SysPanel.AddButton('Setting');
  btn.ImageBig    := Other32;
  btn.DefaultIcon := 1;
  btn.OnClick:=@MenuClick;

  tab := TSPageControl.Create(self);
  tab.parent := SysPanel;
  tab.Align  := alClient;
  tab.OnChange := @TabChange;
  tab.OnClose  := @TabClose;

  statusbar := TStatusBar.Create(self);
  statusbar.parent := self;
  statusbar.Add('',10,saCenter);
  statusbar.Add('',100,saCenter);
  statusbar.Add('',150,saLeft);

  fopendialog:= TSOpenDialog.Create;
  fsavedialog:= TSSaveDialog.Create;

  //SysPanel.Mode := spTop;
  width  := screen.width - screen.width div 6;
  height := screen.height - screen.height div 6;
  factive := nil;
end;

destructor TForm1.Destroy;
begin
  fopendialog.free;
  fsavedialog.free;
  inherited Destroy;
end;

procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: boolean);
begin
  CanClose := CloseAll();
end;

procedure TForm1.TabChange(Sender:TObject);
var
  i: integer;
begin
  factive := TSTab(Sender);
  if factive <> nil then begin
    statusbar.Item(2,TSTab(Sender).path+TSTab(Sender).caption);
    for i:=0 to factive.control.ControlCount-1 do
      if (factive.control.Controls[i] IS TSSynEdit) then begin
        feditor := TSSynEdit(factive.control.Controls[i]).Editor;
        feditor.SetFocus();
        statusbar.Item(0,IntToStr(feditor.CaretX)+':'+IntToStr(feditor.CaretY));
        exit;
      end;
  end else begin
    statusbar.Item(0,'');
    statusbar.Item(1,'');
    statusbar.Item(2,'');
  end;
end;

procedure TForm1.EditorChange(Sender:TObject);
begin
  statusbar.Item(0,IntToStr(TSynEdit(Sender).CaretX)+':'+IntToStr(TSynEdit(Sender).CaretY));
  if factive <> nil then factive.modified := TSynEdit(Sender).Modified;
end;

procedure TForm1.TabClose(Sender:TObject; var done: boolean);
var
  tmp: TSTab;
begin
  tmp  := factive;
  factive := TSTab(Sender);
  done := CloseFile();
  factive := tmp;
end;

procedure TForm1.SetEdit(var se: TSSynEdit; tb: TSTab);
begin
  se := TSSynEdit.Create(self);
  se.visible := false;
  se.align   := alClient;
  se.parent  := tb.control;
  se.visible := true;
  se.Editor.OnChange:=@EditorChange;
  feditor    := se.Editor;
end;

procedure TForm1.NewFile();
begin
  if not tab.visible then tab.visible := true;
  factive := tab.AddNew('Module.mod');
  SetEdit(se,factive);
  feditor.SetFocus();
  statusbar.Item(0,IntToStr(feditor.CaretX)+':'+IntToStr(feditor.CaretY));
end;

procedure TForm1.OpenFile();
var
  i,j: integer;
begin
  if fopendialog.Execute then begin
    if not tab.visible then tab.visible := true;
    j := -1;
    for i:=0 to fopendialog.Files.Count - 1 do begin
      factive := tab.AddTab(fopendialog.Files[i]);
      if not factive.loaded then begin
        SetEdit(se,factive);
        se.LoadFromFile(fopendialog.Files[i]);
      end;
      inc(j);
    end;
    if j > -1 then begin
      feditor.SetFocus();
      statusbar.Item(0,IntToStr(feditor.CaretX)+':'+IntToStr(feditor.CaretY));
    end;
  end;
end;

function TForm1.SaveFileAs(): boolean;
begin
  Result := fsavedialog.Execute;
  if Result then begin
    SDialogs.ShowMessage('Save as '+fsavedialog.FileName);
  end;
end;

function TForm1.SaveFile(): boolean;
var
  i: integer;
begin
  result := factive.path <> '';
  if not Result then
    result := SaveFileAs()
  else if factive.modified then begin
    result := true;
    for i:=0 to factive.control.ControlCount-1 do
      if (factive.control.Controls[i] IS TSSynEdit) then begin
        TSSynEdit(factive.control.Controls[i]).Editor.Lines.SaveToFile(factive.path+factive.caption);
        factive.modified := false;
        exit;
      end;
  end;
end;

function TForm1.SaveAll(): boolean;
var
  tmp : TSTab;
  i: integer;
  done: boolean;
begin
  tmp := factive;
  i   := tab.count - 1;
  result := true;
  while (i >= 0) and result do begin
    factive := tab.GetBy(i);
    result := SaveFile();
    dec(i);
  end;
  factive := tmp;
end;

function TForm1.CloseFile(): boolean;
var
  tmp : TSTab;
begin
  tmp := factive;
  result := not (factive.modified or (factive.path = ''));
  if not result then
    case SDialogs.MessageDlg('Save changed file "'+factive.caption+'" ?',mtInformation,mbYesNoCancel) of
      mbYes    : result := SaveFile();
      mbNo     : result := true;
      mbCancel : result := false;
    end;
  factive := tmp;
  tab.visible := tab.count > 1;
end;

function TForm1.CloseAll(): boolean;
var
  tmp : TSTab;
  i: integer;
begin
  tmp := factive;
  i := tab.count - 1;
  result := true;
  while (i >= 0) and result do begin
    factive := tab.GetBy(i);
    result := CloseFile();
    dec(i);
  end;
  factive := tmp;
end;

procedure TForm1.PrintFile();
begin

end;

procedure TForm1.MenuClick(Sender: TObject; btnindex: integer; menuindex: integer);
begin
  //ShowMessage(IntToStr(btnindex)+' '+IntToStr(menuindex));
  case btnindex of
    0:
      begin
        if menuindex = -1 then menuindex := 0;
        case menuindex of
          0:  NewFile();
          1:  OpenFile();
          2:  SaveFile();
          3:  SaveFileAs();
          4:  SaveAll();
          5:  PrintFile();
          6:  Close;
        end;
      end;
    3:Form2.ShowModal;
  end;
end;

end.

